#!/bin/sh -f

if ! [ -r $1 ] || ! [ -r $2 ]; then exit 2; fi
BASE=`basename $1 .inc`; MAC=$BASE.mac; ASM=$BASE.asm

if [ -e $MAC ]; then mv $MAC $MAC~; fi

cpp -P `\
awk '/^ *#define +[A-Za-z0-9_]+ /{print "-D__"$2"__="$3}' $2 |\
sed -e 's/0[xX]\([0-9A-Fa-f]\+\)/\\$\1/g'\
` -D__INCLUDE__=\"$1\" > $MAC <<'END_OF_FILE'
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;								   ;
;	EEEE	X   X	 PPP		 AA	 CCC	 CCC	   ;
;	E	 X X	 P  P		A  A	C   	C 	   ;
;	EEE	  X	 PPP	===	A  A	C	C	   ;
;	E	 X X	 P		AAAA	C   	C	   ;
;	EEEE	X   X	 P		A  A	 CCC	 CCC 	   ;
;								   ;
;	File : Exp$K3976:AccMain.mac				   ;
;								   ;
;   Control Macro File for Data Taking with K3976         	   ;
;								   ;
;  The Auxiliary CAMAC Crate Controller K3976 equip 40MHz          ;
;  MC68EC030 CPU. This macro package is to support to write Data   ;
;  Taking Program on this machine. 				   ;
;								   ;
;								   ;
;			MACRO Library for Kinetic 3976 		   ;
;								   ;
;			Author 		   M.Iwasaki @ UTMSL	   ;
;					   S.N.Nakamura @ RIKEN	   ;
;			Version number	   1.2.00   95/06/27	   ;
;					   Working Version	   ;
;			Incorporated to RIKEN Version by	   ;
;					   H.Okamura @ Saitama-U   ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; 0) Over View
;
;     Many small to mid-size experiments need fast data aquisition 
;     system in small numbers of CAMAC crates, Fastbus or TKO 
;     system. It depends on the experiment type, though any of them 
;     their data can be handled by a single CAMAC crate. In this 
;     case, it would be best to have a CPU in the crate to handle 
;     the data acuisition. One possibility is to use the product of 
;     KineticSystems Corporation Model 3976 which is a CAMAC 
;     auxiliary Crate Controller equipped with Motolora MC68EC030 
;     40 MHz processor. 
;        This program and its macro packages are to support to write 
;     a Data Acquisition Program (DAQ) on this machine. This Macro 
;     Package reduce the Data Taking Programmer s labor for writing 
;     the program and make it possible to access full capability of 
;     MC68030 without knowing the detail hardware structures and 
;     instructions. To make it possible, a wide variety and high 
;     flexibility Macro routines are prepared aiming that a set of 
;     Macro call is enough to define the various application of DAQ.
;        The ROM Monitor is used only for Self-Test Operation after
;     the power up and the first initialization (F(25)A(0)). After 
;     the first initialization, all RAM region is re-assigned for 
;     interrupt vector, data taking program and data storage region.
;     The successive start application function F(25)A(0) re-defined
;     as Begin RUN (no initialization procedure will start again). 
;     At the end of the first initialization, Two HEX "STATUS" 
;     display shows its RAM size in unit of 100 kB and present its
;     LAM (if host enable the LAM of K3976). It indicates that the 
;     module is waiting for the command from its Host to get into 
;     the data taking phase. Front panel terminal port could not be 
;     used after this initialization.
;        This program can only be compiled with the set of software 
;     in EXP95 named as MAC68030 and ASM68030. Those are the MC68030
;     cross development tools with automatic optimizer on long word 
;     alignment. There are some limitation in instruction and 
;     addressing mode of 68030, though they are enough to write a 
;     data acquisition application on this machine.
;
;
; 1)  Hardware Requirement
;
;     Target machine is KineticSystems Corporation Model 3976. It 
;     could be combined with any CAMAC Crate Controller. The down 
;     loading programs (for the DAQ on K3929) are supported only on 
;     K3929 SCSI Crate Controller at present. 
;        It may be needed output-register (OutReg) to enable and 
;     to disable hardware trigger. Usually, disable the trigger by 
;     putting OutReg output into the veto of the trigger module and 
;     clear it when DAQ starts. To do that clear OutReg just before 
;     Int_Reg or Int_Trg (cf. supported macro) and set OutReg just
;     after those macros. In this way, make it sure not to accept 
;     trigger during DMA. It may also need self-veto into the 
;     trigger until OutReg make response.
;        It would be better to use Interrupt-module (IR). One can 
;     use any modules as a IR if it respond 0 after its clear by 
;     F(9), and if is it possible to select its data bit pattern 
;     after accepting a trigger. When using the module with this 
;     macro, K3976 working in polling mode and according the bit 
;     pattern of the IR, one of 8 routines (Event_0 to Event_7) will
;     be called. Bit 0 is the highest priority and bit 7 is the 
;     lowest.  
;        It is also available to use K3976 front_panel_trigger as 
;     a interrupt source. In this case, Int_Trg macro should be 
;     used. Then K3976 stop and waiting for the trigger. The 
;     argument TP in Int_Trg macro defines which procedure should
;     be call (Event_0 though Event_7). 
;
; 2)  Record Structure and Input Parameters to K3976
;     (snip)
; 3)  Output Parameters from K3976
;     (snip)
; 4)  DAQ procedure and Front Panel "STATUS" LESs
;     (snip)
; 5)  What is to be prepared
;     (snip)
; 6)  Event Structure
;     (snip)
; 7)  Memory Diagram 
;	
;     In this macro, all RAM region is used for DAQ. After the 
;     first initialization, it never come back to use ROM. 
;	
;	     MEMORY MAP OF THE EXP95 SYSTEM ON 3976
;    -------------------------------------------------------------
;	0x5FFFFF	Top of DRAM 4 MB option
;	0x2FFFFF	Top of DRAM 1 MB option
;	0x220000	Start of Data Buffer (in default)
;	0x200400	Start of Data Acquisition Code 
;	0x2003FF	End of interrupt vector table
;	0x200000	Start of DRAM and interrupt vector table
;	0x100000	Start of 3976 registers
;	0x000000	Start of EPROM
;
; 8)  Register Assignment 
;     
;     All registers are assigned to some purpose. The assignment 
;     are summarized as follows.
;
;     constant
;	A3 :  K.CSR 	; Pointer to CSR (CAMAC Status Register)
;	A4 :  K.CCR 	; Pointer to CCR (CAMAC Control Register)
;	A5 :  K.CDR 	; Pointer to CDR (CAMAC Data Register)
;       D5 :  $20	: Sub-Address Step
;	D6 :  $FFFFFF	; Reserved for CAMAC long word MASK
;	D7 :  xxxx	; Switch for Trigger action 
;     variable
;       A0 :  xxxx	: Next_Bank_Pointer 
;	A1 :  xxxx	; Copy data 
;	A2 :  xxxx	; Pointer to Event Starting Point
;	A6 :  xxxx	; Pointer to Data
;       D0 :  xxxx	: Data to get or to put CAMAC (Read/Write)
;			: It will be modified when CAMAC Read 
;			; into D0 
;	D2 :   #I	; Repetition Cycle (loop counter)
;	D3 : (K.CSR)	; CSR Data 	
;	D4 : (K.CCR)	; CCR Data (N, A, F, Block, Q_Wait) 	
;	D1 : (K.CDR) 	; CAMAC Data Buffer
;         Note : If it is needed to use register for adding a new 
;		macro, do it with great care. A1, D0 and D2 are 
;		less dangerous to use. Be carefull that some 
;		macro-call modify those registers. D3, D4, D5 and 
;		D6 can be used according to the assignment. Never 
;		touch to A0, A2, A3, A4, A5, D1 and D6. There is 
;		also some memory region open for user from USR0
;		to USR9 (long-word). 
;
; 9)  Supported Macro
; 
;     A set of macro is preprared in this maro routine to access
;     CAMAC modules. For the simplicity of the code and make the
;     CAMAC access as fast as possible there is no generic CAMAC
;     call is supported. The DAQ programmer need to select one of 
;     the following macro to prepare DAQ. 
;        Those macro can call just like an assembler code of the
;     MC68030. The argument of the macro need to be a certain 
;     number. The default is integer, but you can use hexia decimal
;     by putting $ mark in front (ex. $20 = 32).
;        The description of the typical arguments are given below. 
; 
;   Argument 	     Desctiption
;  ------------------------------------------------------------
;	N	: Module Station Number. 
;	A	: Module Sub-Address. In the case of Scan-Read,
;		  it means the starting Sub-Address of the scan.
;	F	: CAMAC Function
;	Data	: Data to write the module.
;	Max	: Maximum word to be transfered in Block-Read or 
;		  Scan-Read
;	S	: Used in Block-Read or Scan-Read.
;		  1 : Max is the number to be transfered
;		  0 : if D0 < Max, Number of data = D0
;		      if D0 > Max, Number of data = Max.
;	No 	: Station Number of the output register module 
;	Bit 	: Bit Number of output register to make a pulse 
;	Ptr 	: Define internal pointer region (in word). If it 
;		  is 0, then no pointer region is produced.
;	TP  	: Saved as Event-ID in a event_header. Its pattern 
;		  defines the event procedure Event_*.
;	EvNum	: Event number to be accumulated before K3976 makes 
;		  the DMA request
;	PtrNum 	: Define the pointer number for the current memory 
;		  pointer
;		  
;   Basic Procedure without access to Data Buffer
;	NDT	N,A,F	        ; CAMAC Single Action without Data 
;			          Transfer
;	WaitQ	N,A,F		; Wait for Q response for N, A, F
;	W1W	N,A,F,Data	; Write 1 word (Data) to N, A, F
;	W1L	N,A,F,Data	; Write 1 long-word (Data) 
;	W1WD0	N,A,F		; Write Word Data (stored in D0).
;				  Higher word is masked. 
;	WD0	N,A,F		; Write Data (stored in D0).
;				  Same as W1WD0 except it does not 
;				  check Data in D0
;	R1WD0	N,A,F		; Read 1 word and save it to D0
;				  Register 
;	RD0	N,A,F		; Read 1 CAMAC-word and save it 
;				  to D0 Register 
;
;   Basic Procedure to store Data into Data Buffer
;	R1W	N,A,F		; Read 1 word into Data Buffer
;	R1L	N,A,F		; Read 1 long-word into Data Buffer
;	SRW	N,A,F,S,Max	; Sub-Address Scan Word Read into
;				  Data Buffer. Valid in one module.
;	SRL	N,A,F,S,Max	; same as SRL except data size.
;				  (long-word)
;	BRW	N,A,F,S,Max	; Block read (repeat data-read from 
;				  same N, A, F) and save it into 
;				  Data Buffer in word.
;	BRL	N,A,F,S,Max	; same as BRW except data size.
;				  (long-word)
;	BRQW	N,A,F,Max	; Block read on Q-response (repeat 
;				  data-read from same N,A,F until 
;                                 No-Q responce) and store it into 
;				  Data Buffer in word.
;	BRQL	N,A,F,Max	; same as BRQW except data size.
;				  (long-word)
;	Save_D0W		; Save D0 data into Data Buffer 
;				; in word (ignore high word)
;	Save_D0L		; Save D0 data in long-word format
;
;   Module Dependent Procedure
;   (snip)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Begin Initialization Routine ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;



; Global Symbol Definition
	K.LBR	= $100018	; CAMAC LAM Bit Register
	nopr   = __ACC_NOPR__	; address of output reg.
	strtdt = 2		; start data taking bit (bit 1)
	stopdt = 4		; stop data taking bit  (bit 2)
	endisr = __ACC_ENDISR__	; end of interrupt service (bit 9)
; Define program starting Address
	@ = $200400
COM::
IDR::	.Word   0               ; VAX -> 3976 Instruction define reg.
	.Word	0		; dummy for alignment
RSR::	.Word   0               ; 3976 ACC command response
	.Word	0		; dummy for alignment
STR::	.Word   0               ; 3976 ACC status register
	.Word	0		; dummy for alignment
VIR::	.Word   0               ; 3976 -> VAX Instruction define reg.
	.Word	0		; dummy for alignment
VRR::	.Word   0               ; VAX command response
	.Word	0		; dummy for alignment
VSR::	.Word   0               ; VAX status register
	.Word	0		; dummy for alignment
BPT::	.Lword   BPTP            ; Address of buffer prologu table
MPT::	.Lword   MPTP            ; ADDRESS of MEsessage progogu table~
BPTSAV::	
	.Lword   BPTP    ; Address of PBT save area for error recover
outr::	.Word	0
sca::	.Word	0
seqnum::
	.Word	0
	.word	0	; dummy for alignment
MPTP::
; input parameter (with default value)
BankHead::	.Lword	$00000004	; Size of Bank Header (in word)
BankSize::	.Lword	$00000020	; Bank Number to be filled 
LenBuff::	.Lword	$00002000	; Size of 1 Bank = 8kW
DataHead::	.Lword	$00220000	; Pointer to the Data Region 
Command::	.word	$0000		; 1 to continue 0 for stop
Dummy0::	.word	$0000		; dummy
; output parameter
BankNum::	.Lword 	$00000000	; Number of the last filled Bank
EventNum::	.Lword	$00000000	; Event Number from BEGIN
RecNum::	.Lword	$00000000	; Record # = Sum of Bank # from BEGIN
StopRQ::	.Lword	$00000000	; Accept Stop Request from Host
USR0::		.Lword  $00000000	; Memory open for user
USR1::		.Lword  $00000000	; Memory open for user
USR2::		.Lword  $00000000	; Memory open for user
USR3::		.Lword  $00000000	; Memory open for user
USR4::		.Lword  $00000000	; Memory open for user
USR5::		.Lword  $00000000	; Memory open for user
USR6::		.Lword  $00000000	; Memory open for user
USR7::		.Lword  $00000000	; Memory open for user
USR8::		.Lword  $00000000	; Memory open for user
USR9::		.Lword  $00000000	; Memory open for user
COND::		.Lword	$00000000	; CAMAC Status Register @ Error
E_SR::		.word	$0000		; Status Register @ Error
E_PC::		.Lword	$00000000	; Program Counter @ Error 
E_OFF::		.word	$0000		; Vector Offset @ Error
; internal use
BankPntr::	.Lword	$00000000	; Pointer to the current Bank
RunC::		.Lword  $00000000	; Run condition 
;						0: DMA by DMA_Req
;						1: DMA by Data Full
;						2: DAQ phase
RunCB::		.Lword  $00000000	; Run condition Buffer
ACUM::		.Lword  $00000000	; Event to be accumulated
OutReg::	.Lword  $00000000	; Data of Output-Register
LenBuffB::	.Lword	$00000000	; Buffer Length in byte
>::
Begin::
	Move.w	#$2700, SR	; get into supervisor with mask 
				; disable interrupt

; enable instruction cash
	MoveQ.L	#$0019, D0		; enable instruction cash
	MoveC.L	D0, CACR		; set to CACR

; initialize interrupt vector
	MoveA.L	#Error,   A3		; RAM monitor for Error Operation
	MoveA.L #$200008,   A1		; set modify vector Address
	Move.L  #BusErr,   (A1)+ 	; bus error routine (V2)
	Move.L  #AddrErr,  (A1)+ 	; Address error routine (V3)
	Move.L  A3,  (A1)+	; illegal instruction (V4)
	Move.L  A3,  (A1)+	; zero dev (V5)
	Move.L  A3,  (A1)+	; chk, chk2 (V6)
	Move.L  A3,  (A1)+	; trapcc trapv, cctrapcc (V7)
	Move.L  A3,  (A1)+	; privilege exception (V8)
	MoveA.L #$200034,   A1
	Move.L  A3,  (A1)+	; copro protocol (V13)
	Move.L  A3,  (A1)+	; format error (V14)
	Move.L  A3,  (A1)+	; undefined interrupt (V15)
	MoveA.L #$200060,   A1
	Move.L  A3,  (A1)+	; Spurious Interrupt (V24)
	Move.L  A3,  (A1)+	; reserved (V25)
	Move.L  A3,  (A1)+	; LAM is not supported (V26)
	Move.L  #HostRQ,   (A1)+	; CAMAC command (V27)
	Move.L  A3,        (A1)+	; CAMAC Done (V28)
	Move.L  #Trig,     (A1)+	; Front Panel Trigger (Timer V29)
	Move.L  A3,        (A1)+	; DUART (Front Panel Terminal) (V30)
	Move.L  #Begin,    (A1)+	; Begin Request From Host (V31)
	MoveA.L	#$200080,   A1	; set to Vector 32
	Move.L  A3,  (A1)+	; Trap #0 (V32)
	Move.L  A3,  (A1)+	; Trap #1 (V33)
	Move.L  A3,  (A1)+	; Trap #2 (V34)
	Move.L  A3,  (A1)+	; Trap #3 (V35)
	Move.L  A3,  (A1)+	; Trap #4 (V36)
	Move.L  A3,  (A1)+	; Trap #5 (V37)
	Move.L  A3,  (A1)+	; Trap #6 (V38)
	Move.L  A3,  (A1)+	; Trap #7 (V39)
	Move.L  A3,  (A1)+	; Trap #8 (V40)
	Move.L  A3,  (A1)+	; Trap #9 (V41)
	Move.L  A3,  (A1)+	; Trap #A (V42)
	Move.L  A3,  (A1)+	; Trap #B (V43)
	Move.L  A3,  (A1)+	; Trap #C (V44)
	Move.L  A3,  (A1)+	; Trap #D (V45)
	Move.L  A3,  (A1)+	; Trap #E (V46)
	Move.L  A3,  (A1)+	; Trap #F (V47)
	MoveA.L	#$2000C0,   A1	; set to Vector 48
	Move.L  A3,  (A1)+	; FPCP (V48)
	Move.L  A3,  (A1)+	; FPCP (V49)
	Move.L  A3,  (A1)+	; FPCP (V50)
	Move.L  A3,  (A1)+	; FPCP (V51)
	Move.L  A3,  (A1)+	; FPCP (V52)
	Move.L  A3,  (A1)+	; FPCP (V53)
	Move.L  A3,  (A1)+	; FPCP (V54)
	MoveA.L  #$2000E0,   A1	; set to vector 56
	Move.L  A3,  (A1)+	; MMU configuration error (V56)
; register definition
;	K.CSR 	= $100000	; CAMAC Status Register
;	K.CCR 	= $100004	; CAMAC Function Register
;	K.CDR 	= $100008	; CAMAC Data Register
	MoveA.L	#$100000, A3	; set CSR (CAMAC Status Register)
	MoveA.L	#$100004, A4	; set CCR (CAMAC Control Register)
	MoveA.L	#$100008, A5	; set CDR (CAMAC Data Register)
	MoveQ.L	#$20, D5	; Set Sub-Address Step
	Move.L	#$FFFFFF, D6	; D6 is reserved for CAMAC long word MASK
;
; Initialize Buffer area
;
	Bsr	clear		; clear buffer
;
; display memory size 
;
	AddQ.L	#$8, SP		; release stack 
	Move.L	SP, D3		; get memory end to D3
	MoveA.L	(DataHead), SP	; initialize Stack Pointer
	SubI.L	#$200000, D3	; get memory size
	Move.W	#$100, D3	; Enable CAMAC_Interrupt (F25A1)
	Move.L	D3, (A3)	; display memory size
;
; wait for interrupt
;
	Move.W	#$2200,SR	; set interrupt level=2
Loop1::	bra	Loop1		; wait for interrupt


BusErr::
	Move.W	#$2700, SR	; accept no interrupt
	Move.L	(A3), D3	; get CAMAC Status Register
	AndI.L	#$FFFF, D3	; clear display word
	OrI.L	#$E10000, D3	; set display word = E1
	Bra	BAErr		; Branch to Branch & Address Error
AddrErr::
	Move.W	#$2700, SR	; accept no interrupt
	Move.L	(A3), D3	; get CAMAC Status Register
	AndI.L	#$FFFF, D3	; clear display word
	OrI.L	#$E20000, D3	; set display word = E2
BAErr::
	Move.L	D3, (A3)	; display error code
	MoveA.L	#COND, A1	; get Address to save Error condition
	Move.W	(SP)+, (A1)+	; save SR to E_SR
	Move.W	(SP)+, (A1)+	; save PC.msw to E_PC.msw
	Move.W	(SP)+, (A1)+	; save PC.lsw to E_PC.lsw
	Move.W	(SP),  (A1)	; save offset to E_OFF
	MoveA.L	(DataHead), SP	; reset stack pointer
	Move.L	#$1, (K.LBR)	; set LAM of K3976
	Stop	#$2700


Error::
	Move.W	#$2700, SR	; accept no interrupt
	Move.L	(A3), D3	; get CAMAC Status Register
	MoveA.L	#COND, A1	; get Address to save Error condition
	Move.L	D3, (A1)+	; save CSR to COND
	AndI.L	#$FFFF, D3	; clear display word
	Move.W	(SP)+, (A1)+	; save SR to E_SR
	Move.W	(SP)+, (A1)+	; save PC.msw to E_PC.msw
	Move.W	(SP)+, (A1)+	; save PC.lsw to E_PC.lsw
	Move.W	(SP),  (A1)	; save offset to E_OFF
	Move.W	(SP)+, D6	; get format
	AndI.W	#$F000, D6	; mask offset
	CmpI.W	#$0000, D6	; check error type 
	Beq	Err1		; if type 3 error
	CmpI.W	#$1000, D6	; check error type 
	Beq	Err2		; if type 4 error
	CmpI.W	#$2000, D6	; check error type 
	Beq	Err3		; if type 5 error
	CmpI.W	#$9000, D6	; check error type 
	Beq	Err4		; if type 6 error
	OrI.L	#$E70000, D3	; set display word = E7
	Bra	ErrEnd
Err1::
	OrI.L	#$E30000, D3	; set display word = E3
	Bra	ErrEnd
Err2::
	OrI.L	#$E40000, D3	; set display word = E4
	Bra	ErrEnd
Err3::
	OrI.L	#$E50000, D3	; set display word = E6
	Bra	ErrEnd
Err4::
	OrI.L	#$E60000, D3	; set display word = E6
ErrEnd::
	MoveA.L	(DataHead), SP	; reset stack pointer
	Move.L	D3, (A3)	; display error code
	Move.L	D3, (COND)	; save CAMACSR to COND
	Move.L	#$1, (K.LBR)	; set LAM of K3976
	Stop	#$2700


HostRQ::				; (F25A1)
	Move.W	#$2700, SR	; Accept no interrupt
	Move.W	(idr),D7	; Read command register
	;
	; Analyse command and call execute routine
	;
	Cmp.W	#1,D7		; check command 1=Clear
	Bne	cmd2		; if not, next command
	Bsr	clear		; clear buffer
	Bra	cmdend		; goto command end
cmd2::	Cmp.W	#2,D7		; check command 2=change buffer
	Bne	cmd3		; if not, next command
	Bsr	chgbuf		; clear buffer
	Bra	cmdend		; goto command end
cmd3::	Cmp.W	#4,D7		; check command 3=start data acquisition
	Bne	cmd4		; if not, next command
	Bsr	starta		; clear buffer
	Bra	cmdend		; goto command end
cmd4::	Cmp.W	#8,D7		; check command 4=stop data acquisition
	Bne	cmd5		; if not, next command
	Bsr	stopa		; clear buffer
	Bra	cmdend		; goto command end
cmd5::	Cmp.W	#$10,D7		; check command 5=abort DMA
	Bne	errcmd		; if not, next command
	Bsr	enddma		; clear buffer
	Bra	cmdend		; goto command end
	;
	; Set Command response (done or error) and return
	;
errcmd::
	Move.W	#$8000,(rsr)	; set error bit of command response
	Move.W	(str),D1	; get status reg. value
	OrI.W	#$0080,D1	; set command error bit of status reg.
	AndI.W	#$FFFE,D1	; remove command done bit
	Move.W	D1,(str)	; store status reg.
	MoveQ.L	#$E,D0		; set error code for status display 
	Bra	cmddsp		; jump display routine
cmdend::
	Move.W	#1,(rsr)	; set done successful bit
	Move.W	(str),D1	; get status reg. value
	OrI.W	#1,D1		; set command done bit of status reg.
	AndI.W	#$FF7F,D1	; remove command error bit
	Move.W	D1,(str)	; store status reg.
	MoveQ.L	#5,D2		; set max_loop count
	Clr.L	D0		; reset count
lp1::	AddQ.L	#1,D0		; count up
	Lsr.L	#1,D7		; shift left
	DBcs	D2,lp1		; if Carry is set, exit
cmddsp::
	Move.L	(A3), D3	; get CAMAC Status Register
	AndI.L	#$FFFF, D3	; clear display word
	Swap	D0		; bring lower word to upper word
	Or.L	D0,D3		; set command to display
	OrI.L	#$C00000, D3	; set display word = C?
	Move.L	D3,(A3)		; write Status Register
	Rte			; return from interrupt

;
; Execute each command
;
clear::
	Bsr	clrb1		; clear buffer 1
	Bsr	clrb2		; clear buffer 2
	Bsr	setb1		; set buffer1 write mode
	Clr.L	(bcbf)		; clear bcbf
	Move.W	(str),D1	; get status
	AndI.W	#$FE3F,D1	; clear error bits
	Move.W	D1,(str)	; store
	Rts			; return

chgbuf::
	Bsr	exchbf		; exchange buffer
	Rts			; return

starta::
	Bsr	intmdl		; CAMAC module initialization
	Clr.W	(seqnum)	; Clear event sequential number
	Clr.W	(sca)
	Clr.W	(outr)
	strtwd = endisr+strtdt	; word data for output reg
	W1W	nopr,0,17,strtwd
				; write start bit and done bit to output reg.
	Move.L	(A3),D1		; Get CAMAC status reg.
	OrI.W	#$4040,D1	; Enable trigger (Set bit6 and bit14)	
	Move.L	D1,(A3)		; Store value
	Rts			; return to main

stopa::
	Bsr	exchbf		; Change buffer
	Move.L	(A3),D1		; Get CAMAC status reg.
	AndI.W	#$BFBF,D1	; disable trigger (Clear bit6 and bit14)
	Move.L	D1,(A3)		; Store value
	W1W	nopr,0,17,stopdt	; write to output reg.
	Rts

enddma::
	Move.W	(str),D1	; Get str value
	AndI.W	#$FFFB,D1	; clear dma flag (bit2)
	Move.W	D1,(str)	; Store value
	Rts

;
; System constant definition
;
	header = __ACC_EVT_ID__	; event header (lower 8-bit is event counter)
	maxevl = __ACC_MAXEVL__*2	; maximum event length in byte
	scasiz = __SCL_NUM_DATA__*4	; size of scaler area in byte

;
; Routines provided by user
;

#include __INCLUDE__

;
; Exchange buffer
;
exchbf::
	;
	; Block Header
	;
	Move.L	(bcbw),A1	; get BCB address
	Move.L	(8,A1),D0	; get buffer length
	Move.L	(A1),A1		; get buffer start address
	Clr.W	(A1)		; set block ID = 0
	SubI.L	#scasiz,D0	; calc. buffer limit
	Move.W	D0,(4,A1)	; store buffer limit
	Move.L	A6,D0		; get current address
	Sub.L	A1,D0		; calc. effective buffer length
	Move.W	D0,(6,A1)	; store effective buffer length
	Move.W	(seqnum),D0	; get event counter
;	AndI.W	#$0FFF,D0	; clear high 4-bit
	Move.W	D0,(2,A1)	; store (crate address, flag, event counter)

	MoveA.L (bcbw),A1	; Get BCB address
	Move.L	A6,(16,A1)	; write current pointer
	;
	; change buffer
	;
	Clr.W	(seqnum)	; Clear event sequential number
	Tst.W	(bcbf)		; Is there any empty buffer?
	beq	chbf1		; yes

	Bsr	waitbf		; waiting for buffer free
	Bcs	chbf11		; error end

chbf1::	Move.L	(bcbw),D0	; Get current write buffer
	Bne	chbf2		;
	Bsr	errbuf		; error

chbf2:: Move.L	D0,(bcbf)	;
	Cmp.L	#1,(bcbw1)	; Buffer 1?
	bne	chbf3		;
	;
	; previous buffer is no.1
	;
	Move.W	#2,(vir)	; set Buffer 1 full flag
	Move.L	(bcb1),(bcbf)	; set buffer full pointer
	Bsr	setb2		; set buffer2 for next write
	Bra	chbf10		;
	;
	; previous buffer is no.2
	;
chbf3::	Move.W	#4,(vir)	; set Buffer 2 full flag
	Move.L	(bcb2),(bcbf)	; set buffer full pointer
	Bsr	setb1		; set buffer1 for next write
	;
	; Tell to the host that the buffer is full
	;
chbf10::
	Move.W	(str),D1	; Get str value
	OrI.W	#4,D1		; Set dma flag
	Move.W	D1,(str)	; Store value
	Move.L	#1,(k.lbr)	; Generate LAM
chbf11::
	Rts
	;
	; There is now no room for next buffer
	; waiting for buffer free
	;
waitbf::
	Move.L	#100,D1		; set loop counter
wait1::	Move.L	#1000,D0	; set loop counter
wait2::	DBf	D0,wait2	; wait
	Tst.L	(bcbf)		; Check bcbf
	Beq	wait10		; buffer is ready
	DBf	D1,wait1	; repeat until timeout
	Bsr	errbuf		; Error buffer
	OrI.B	#1,CCR		; Set carry (bit0)
	bra	wait11		;
wait10::
	AndI.B	#$FE,CCR	; Clear carry (bit0)
Wait11::
	Rts			; return
	;
	; Fatal error in buffer manipulation
	;
errbuf::
	Move.W	(str),D1	; Get str value
	OrI.W	#$40,D1		; Set buffer error bit
	Move.W	D1,(str)	; Store value
	Bsr	clrb1		; clear buffer1
	Bsr	setb1		; Set buffer1 for next write
	Clr.L	(bcbf)		; Set empty
	Move.W	#$100,(vir)	; set buffer error flag
	Rts			; return
;
; Clear Buffer
;
clrb1::	MoveA.L	(bcb1),A0	; Set buffer control block 1
	bra	clrb
clrb2::	MoveA.L	(bcb2),A0	; Set buffer control block 2
clrb::	MoveA.L	(A0),A1		; Get buffer address
	Move.L	(8,A0),D0	; Length of the buffer in byte
	Asr.L	#2,D0		; Length of the buffer in Lword
	SubQ.L	#1,D0		; Decrement the count
clr1::	Clr.L	(A1)+		; Clear buffer
	DBf	D0,clr1		; Loop
	Clr	(12,A0)		; Clear Buffer status register
	Move.L	(A0),(16,A0)	; Copy top address to write pointer
	Move.L	(A0),(20,A0)	; Copy top address to read pointer
	Rts
;
; Set Buffer 1 or 2
;
setb1::	MoveA.L	(bcb1),A0	; Get address of buffer control block1
	Move.L	#1,(bcbw1)	; buffer number for write
	Move.L	#$B10000,D0	; Set Buffer 1 use
	Bra	setb
setb2::	MoveA.L	(bcb2),A0	; Get address of buffer control block1
	Move.L	#2,(bcbw1)	; buffer number for write
	Move.L	#$B20000,D0	; Set Buffer 2 use

setb::	Move.L	A0,(bcbw)	; BPT for write
	Move.L	(A0),D3		; Get address of the buffer
	AddI.L	#8,D3		; Make Space for Header Area
	Move.L	D3,A6		; Store pointer
	Move.L	A6,(16,A0)	; Copy top address of the buffer
	Move.W	#2,(12,A0)	; Status <= Write mode
	Move.L	(A3), D3	; get CAMAC Status Register
	AndI.L	#$FFFF, D3	; clear display word
	Or.L	D0,D3		; Set display code (D?)
	Move.L	D3,(A3)		; Store to csr
	Rts
;
;       Work Area
;

;
; Buffer prologue table
;
BPTP::
BCBW::	.Lword	0		; Address of BCB for current write
BCBW1::	.Lword	0		; Buffer no. for current write
BCBR::	.Lword	0		; Address of BCB for current read
BCBR1::	.Lword	0		; Buffer no. for current read
BCBF::	.Lword	0		; Address of BCB for data full
BCBF1::	.Lword	0		; Buffer no. for data full
BCB1::	.Lword	BCBA1		; Address of Buffer control Block 1
BCB2::	.Lword	BCBA2		; Address of Buffer control Block 2
	.word	0,0,0,0,0,0	; (Reserved)

	bufl =  $4000		; 8k word
	buff1 = $280000		; Buffer1 address
	buff2 = $288000		; Buffer2 address
;
; Buffer control block 1
;
BCBA1::	.Lword	$280000		; Address of the buffer
	.Lword	$284000		; Limit of the buffer
	.Lword	$4000		; Length of the buffer
	.word	0		; Status of buffer
	.word	0		; Acess Inter-lock
	.Lword	$280000		; Write data pointer
	.Lword	$280000		; Read data pointer
	.Lword	0		; (reserved)
;
; Buffer control block 2
;
BCBA2::	.Lword	$288000		; Address of the buffer
	.Lword	$28C000		; Limit of the buffer
	.Lword	$4000		; Length of the buffer
	.word	0		; Status of buffer
	.word	0		; Acess Inter-lock
	.Lword	$288000		; Write data pointer
	.Lword	$288000		; Read data pointer
	.Lword	0		; (reserved)




	.MACRO	NAFset, N, A, F, BLK, QRP
	.if GE, N - 32
	  .ermsg  N, exceed, 31
	.endif
	.if GE, A - 16
	  .ermsg  A, exceed, 15
	.endif
	.if GE, F - 32
	  .ermsg  A, exceed, 31
	.endif
	.if GE, BLK - 2
	  .ermsg  BLK, exceed, 1
	.endif
	.if GE, QRP - 2
	  .ermsg  QRP, exceed, 1
	.endif
	NAF = BLK*$8000 + QRP*$4000 + N*$200 + A*$20 + F
	.ENDM





	.MACRO	WaitDone
	Move.L	#$20,D1		; max trial number
WD:	Move.L  (A3), D3	; read status reg.
	RoR.L    #4,  D3	; Check Done bit
	Bcs	 WD1		; If  Done then exit
	SubQ.L	#1,D1		; decriment counter
	Bne	WD		; retry
WD1:
	.ENDM




;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; No Data Transfer Routines ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;



; CAMAC Single Action with No Data Transfer
	.MACRO	NDT, N,A,F	; No data transfer 
	NAFset	N, A, F, 0, 0	; BLK = 0, QRP = 0 
	Move.L	#NAF, (A4)	; transfer NAF code to CFR (go!)
	WaitDone
	.ENDM



; CAMAC Single Action with No Data Transfer and Wait Q_Response
	.MACRO	WaitQ, N,A,F	; No data transfer 
	NAFset	N, A, F, 0, 1	; BLK = 0, QRP = 1 
	Move.L	#NAF, (A4)	; transfer NAF code to CFR (go!)
	WaitDone
	.ENDM



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;    Write Data Routines    ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;



; Write 1 word (Data) to N, A, F
	.MACRO	W1W, N,A,F,Data	 ; Write 1 word
;	F	:CAMAC function
;	N, A	:station Address and Sub-Address
;	Data	:data 
	NAFset,	N, A, F, 0, 0	 ; BLK = 0, QRP = 0
	.if	GT, Data - $FFFF
	  .ERMSG   Exceed, Word, Limit
	.endif
	Move.L  #Data, (A5)	 ; Write data to CDR (one word)
	Move.L	#NAF,  (A4)	 ; transfer NAF code to CFR (go!)
	WaitDone
	.ENDM



; Write 1 long-word (Data) to N, A, F
	.MACRO	W1L, N,A,F,Data	; write 1 long word
;	F	:CAMAC function
;	N, A	:station Address and Sub-Address
;	Data	:data 
	.if	GT, Data - $FFFFFF
	  .ERMSG   Exceed, CAMAC-Word, Limit
	.endif
	NAFset	N, A, F, 0, 0	; BLK = 0, QRP = 0
	Move.L  #Data, (A5)	 ; Write data to CDR (one word)
	Move.L	#NAF, (A4)	; transfer NAF code to CFR (go!)
	WaitDone		; Wait for CAMAC Cycle
	.ENDM




; Write Word Data (stored in D0) to  N, A, F 
	.MACRO	W1WD0, N,A,F	; Write CAMAC word from D0 
;	F	:CAMAC function
;	N, A	:station Address and Sub-Address
	NAFset	N, A, F, 0, 0	; BLK = 0, QRP = 0
	AndI.L	#$FFFF, D0	; Get Word Data
	Move.L	D0,    (A5)	; Write data to CDR (one word)
	Move.L	#NAF,  (A4)	; transfer NAF code to CFR (go!)
	WaitDone		; Wait for CAMAC Cycle
	.ENDM



; Write Data (stored in D0) to  N, A, F 
	.MACRO	WD0, N,A,F	; Write CAMAC word from D0 
;	F	:CAMAC function
;	N, A	:station Address and Sub-Address
	NAFset	N, A, F, 0, 0	; BLK = 0, QRP = 0
	Move.L	D0, (A5)	; Write data to CDR (one word)
	Move.L	#NAF, (A4)	; transfer NAF code to CFR (go!)
	WaitDone		; Wait for CAMAC Cycle
	.ENDM



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;    Read Data Routines     ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;



; Read 1 word from N, A, F and save it to Data Buffer
	.MACRO	R1W, N,A,F	;read 1 word
;	F	:CAMAC function
;	N, A	:station Address and Sub-Address
;	D	:data 
	NAFset	N, A, F, 0, 0	; BLK = 0, QRP = 0
	Move.L	#NAF, (A4)	; Transfer NAF code to CFR (go!)
	WaitDone		; Wait for CAMAC Cycle
	Move.L  (A5), D1	; Read Data from CDR 
	Move.W	D1, (A6)+	; Word Data Transfer
	.ENDM



; Read 1 long-word from N, A, F and save it to Data Buffer
	.MACRO	R1L, N,A,F	;read 1 long word
;	F	:CAMAC function
;	N, A	:station Address and Sub-Address
;	D	:data 
	NAFset 	N, A, F, 0, 0	; BLK = 0, QRP = 0
	Move.L	#NAF, (A4)	; transfer NAF code to CFR (go!)
	WaitDone		; Wait for CAMAC Cycle
	Move.L  (A5), D1	; Read Data from CDR 
	And.L	D6, D1		; Get CAMAC-Word
	Move.W	D1, (A6)+	; Word Data Transfer
	.ENDM



; Read 1 word from N, A, F and save it to D0 Register
	.MACRO	R1WD0, N,A,F	; Read CAMAC word to D0
;	F	:CAMAC function
;	N, A	:station Address and Sub-Address
	NAFset	N, A, F, 0, 0	; BLK = 0, QRP = 0
	Move.L	#NAF, (A4)	; transfer NAF code to CFR (go!)
	WaitDone		; Wait for CAMAC Cycle
	Move.L	(A5), D1	; Write data to CDR 
	Clr.L	 D0		; Clear D0 Reg.
	Move.W	 D1, D0		; Get Lower Word
	.ENDM



; Read 1 CAMAC-word from N, A, F and save it to D0 Register
	.MACRO	RD0, N,A,F	;read 1 CAMAC-word to D0
;	F	:CAMAC function
;	N, A	:station Address and Sub-Address
	NAFset	N, A, F, 0, 0	; BLK = 0, QRP = 0
	Move.L	#NAF, (A4)	; transfer NAF code to CFR (go!)
	WaitDone		; Wait for CAMAC Cycle
	Move.L	(A5), D0	; Read data from CDR 
	And.l	D6, D0
	.ENDM



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Address Scan Read Routines ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;



; Sub-Address Scan Word Read
	.MACRO	SRW,N,A,F,S,I	; Scan read (word) in a CAMAC module
;	F	:CAMAC function
;	N A	:station Address and the first Sub-address to be read
;	S    	:Switch; S=0 if D0-register is used to be word MAX counter
;	I    	:Number of Iteration 
;		: S=0, D0<I ... Max word = D0
;		: S=0, D0>I ... Max word = I
;		: S=1,      ... Max word = I
	.if GT, I - 16
	  .ermsg  Max, Exceeded, SubA, (16)
	.endif
	.if EQ, I
	  .ermsg  Scan, Read, for, 0, word
	.endif
	NAFset 	N, A, F, 0, 0	; BLK = 0, QRP = 0
	Move.L	#NAF, D4	; Save NAF into D4
	Move.L   D4, (A4)	; Transfer NAF code to CFR (go!)
	.if NE, S		; If fixed word read mode requested
	  .repeat I-1		; Repeat CAMAC cycle I - 1
	    Add.L     D5,  D4	; Increase SubAddress (A)
	    WaitDone		; Wait for CAMAC Cycle
	    Move.L   (A5), D1	; Read data from CDR 
	    Move.L    D4, (A4)	; Next transfer NAF code to CFR (go!)
	    Move.W    D1, (A6)+ ; Data Store to Buffer
	  .endrepeat		; 
	  WaitDone		; Wait for CAMAC Cycle
	  Move.L   (A5), D1	; Read data from CDR 
	  Move.W    D1, (A6)+   ; Data Store to Buffer
	.endif
	.IF EQ, S		; If D0 repeat mode requested
	  Move.L     #I,  D2	; Set max. word count to D2
	    Cmp.L    D2,  D0	; Compare D0 and #I
	    Bge	     SRW1	; D0 >= I then loop #I (D2)
	  Move.L     D0,  D2	; (D0 < I)
SRW1:	  Add.L      D5,  D4	; Increase SubAddress (A)
	  SubQ.L     #1,  D2	; Subtract already done cycle
SRW2:	    WaitDone		; Wait for CAMAC Cycle
	    Move.L  (A5), D1	; Read data from CDR 
	    Move.L   D4, (A4)	; Next transfer NAF code to CFR (go!)
	    Move.W   D1, (A6)+  ; Data Store to Buffer
	    Add.L    D5,  D4	; Increase SubAddress (A)
	    SubQ.L  #$1,  D2	; Loop Count
	  Bne SRW2
	  WaitDone		; Wait for CAMAC Cycle
	  Move.L  (A5), D1	; Read data from CDR 
	  Move.W   D1, (A6)+    ; Data Store to Buffer
	.ENDIF
	.ENDM



; Sub-Address Scan Long-Word Read
	.MACRO	SRL,N,A,F,S,I	; Scan read (word) in a CAMAC module
;	F	:CAMAC function
;	N A	:station Address and the first Sub-address to be read
;	S    	:Switch; S=0 if D0-register is used to be word MAX counter
;	I    	:Number of Iteration 
;		: S=0, D0<I ... Max word = D0
;		: S=0, D0>I ... Max word = I
;		: S=1,      ... Max word = I
	.if GT, I - 16
	  .ermsg  Max, Exceeded, SubA, (16)
	.endif
	.if EQ, I
	  .ermsg  Scan, Read, for, 0, word
	.endif
	NAFset 	N, A, F, 0, 0	; BLK = 0, QRP = 0
	Move.L	#NAF, D4	; Save NAF into D4
	Move.L	D4, (A4)	; Transfer NAF code to CFR (go!)
	.if NE,S		; If fixed word read mode requested
	  .repeat I-1
	    Add.L     D5,  D4	; Increase SubAddress (A)
	    WaitDone		; Wait for CAMAC Cycle
	    Move.L   (A5), D1	; Read data from CDR 
	    Move.L    D4, (A4)	; Next transfer NAF code to CFR (go!)
	    And.L     D6,  D1	; Get CAMAC-Word
	    Swap      D1
	    Move.L    D1, (A6)+ ; Data Store to Buffer
	  .endrepeat
	  WaitDone		; Wait for CAMAC Cycle
	  Move.L   (A5), D1	; Read data from CDR 
	  And.L     D6,  D1	; Get CAMAC-Word
	  Swap      D1
	  Move.L    D1, (A6)+   ; Data Store to Buffer
	.endif
	.IF	EQ,S		; If D0 repeat mode requested
	  Move.L     #I,  D2	; Set max. word count to D2
	    Cmp.L    D2,  D0	; Compare D2 and D0
	    Bge	     SRL1	; D0 >= I then go SRL1
	  Move.L     D0,  D2	; (D0 < I)
SRL1:	  Add.L      D1,  D4	; Increase SubAddress (A)
	  SubQ.L     #1,  D2	; Subtract already done cycle
SRL2:	    WaitDone		; Wait for CAMAC Cycle
	    Move.L  (A5), D1	; Read data from CDR 
	    Move.L   D4, (A4)	; Next transfer NAF code to CFR (go!)
	    And.L    D6,  D1	; Get CAMAC-Word
	    Swap     D1
	    Move.L   D1, (A6)+  ; Data Store to Buffer
	    Add.L    D5,  D4	; Increase SubAddress (A)
	    SubQ.L  #$1,  D2	; Loop Count
	  Bne SRL2
	.ENDIF
	.ENDM



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;    Block Read Routines     ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;



; Block read (repeat data-read from same N,A,F)	in word
	.MACRO	BRW,N,A,F,S,I	; Block Read (word) 
;	F	:CAMAC function
;	N, A	:station Address and Sub-Address
;	S    	:Switch
;		    S=0 : D0 is taken as word MAX count	
;		    S=1 : I is the word count
;	I    	:Number of Iteration 
;		: S=0, D0<I ... Max word = D0
;		: S=0, D0>I ... Max word = I
;		: S=1,      ... Max word = I
	.if EQ, I
	  .ermsg  Block, Read, for, 0, word
	.endif
	NAFset 	N, A, F, 1, 0	; BLK = 1, QRP = 0
	Move.L	#NAF, (A4)	; Transfer NAF code to CFR (go!)
	.IF	NE,S		; If fixed word read mode requested
	  Move.L     #I,   D2	; Set max. word count to D0
BRW1:	  WaitDone		; Wait for CAMAC Cycle Done
	    Move.L  (A5),  D1	; Read data and Generate next cycle (go!)
	    Move.W   D1,  (A6)+	; Send word to Data Buffer
	    SubQ.L  #$1,   D2   ; Loop count
	  Bne BRW1		; Do until end
	.ENDIF
	.IF	EQ,S		; if D0 repeat mode requested
	  Move.L     #I,   D2	; set max. word count to D2
	    Cmp.L    D2,   D0	; Compare D0 and #I
	    Bge	     BRW2	; D0 >= I then loop #I (D2)
	  Move.L     D0,   D2	; (D0 < I)
BRW2:	  WaitDone		; Wait for CAMAC Cycle Done
	    Move.L  (A5),  D1	; Read data from CDR 
	    Move.W   D1,  (A6)+ ; Send word to Data Buffer
	    SubQ.l  #$1,   D2   ; Loop count
	  Bne BRW2		; Do until end
	.ENDIF
	.ENDM



; Block read (repeat data-read from same N,A,F)	in long word
	.MACRO	BRL,N,A,F,S,I	; Block Read (long word) 
;	F	:CAMAC function
;	N, A	:station Address and Sub-Address
;	S    	:Switch
;		    S=0 : D0 is taken as long-word count
;		    S=1 : I is the long-word count
;	I    	:Number of Iteration 
;		: S=0, D0<I ... Max word = D0
;		: S=0, D0>I ... Max word = I
;		: S=1,      ... Max word = I
	.if EQ, I
	  .ermsg  Block, Read, for, 0, word
	.endif
	NAFset 	N, A, F, 1, 0	; BLK = 1, QRP = 0
	Move.L	#NAF, (A4)	; Transfer NAF code to CFR (go!)
	.IF	NE,S		; If fixed long-word read mode requested
	  Move.L     #I,   D2	; Set max. long-word count to D0
BRL1:	  WaitDone		; Wait for CAMAC Cycle Done
	    Move.L  (A5),  D1	; Read data and Generate next cycle (go!)
	    And.L    D6,   D1   ; Mask Most Significant byte
	    Swap     D1
	    Move.L   D1,  (A6)+	; Send long-word to Data Buffer
	    SubQ.L  #$1,   D2   ; Loop count
	  Bne BRL1		; Do until end
	.ENDIF
	.IF	EQ,S		; if D0 repeat mode requested
	  Move.L     #I,   D2	; set max. long-word count to D2
	    Cmp.L    D2,   D0	; Compare D0 and #I
	    Bge	     BRL2	; D0 >= I then loop #I (D2)
	  Move.L     D0,   D2	; (D0 < I)
BRL2:	  WaitDone		; Wait for CAMAC Cycle Done
	    Move.L  (A5),  D1	; Read data from CDR 
	    And.L    D6,   D1   ; Mask Most Significant byte
	    Swap     D1
	    Move.L   D1,  (A6)+ ; Send long-word to Data Buffer
	    SubQ.l  #$1,   D2   ; Loop count
	  Bne BRL2		; Do until end
	.ENDIF
	.ENDM



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Q-Stop Block Read Routines ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;



; Block read on Q-response (repeat data-read from same N,A,F) in word
	.MACRO	BRQW,N,A,F,I	; Block Read on Q-Stop (word) 
;	F	:CAMAC function
;	N, A	:station Address and Sub-Address
;	I    	:Number of Iteration 
	NAFset 	N, A, F, 1, 0	; BLK = 1, QRP = 0
	Move.L	#NAF, (A4)	; Transfer NAF code to CFR (go!)
	  Move.L     #I,   D2	; Set max. long-word count to D2
RQW1:	  WaitDone		; Wait for CAMAC Cycle Done
	    Move.L  (A5),  D1	; Read data and Generate next cycle (go!)
	    RoL.L    #4,   D3	; check Q_response
	    Bcs	    RQW2	; if no-Q Stop
	    Move.W   D1,  (A6)+	; Send long-word to Data Buffer
	    SubQ.L  #$1,   D2   ; Loop count
	  Bne RQW1		; Do until end
RQW2:
	.ENDM



; Block read on Q-response (repeat data-read from same N,A,F) in long-word
	.MACRO	BRQL,N,A,F,I	; Block Read on Q-Stop (long word) 
;	F	:CAMAC function
;	N, A	:station Address and Sub-Address
;	I    	:Number of Iteration 
	NAFset 	N, A, F, 1, 0	; BLK = 1, QRP = 0
	Move.L	#NAF, (A4)	; Transfer NAF code to CFR (go!)
	  Move.L     #I,   D2	; Set max. long-word count to D2
BRQL1:	  WaitDone		; Wait for CAMAC Cycle Done
	    Move.L  (A5),  D1	; Read data and Generate next cycle (go!)
	    RoL.L    #4,   D3	; check Q_response
	    Bcs	    BRQL2	; if no-Q Stop
	    And.L    D6,   D1   ; Mask Most Significant byte
	    Swap     D1
	    Move.L   D1,  (A6)+	; Send long-word to Data Buffer
	    SubQ.L  #$1,   D2   ; Loop count
	  Bne BRQL1		; Do until end
BRQL2:
	.ENDM

; Block word-read on Q-response with module header
	.MACRO	BRHQW,N,A,F,I	; Block Read on Q-Stop (long word) 
;	F	:CAMAC function
;	N, A	:station Address and Sub-Address
;	I    	:Number of Iteration 
	Headset N		; make header word
	Move.W	#header,(A6)+	; write module header
	Move.L	A6,A0		; store header address
	BRQW	N,A,F,I		; Q-stop read
	Move.L	A6,D0		; copy current pointer
	Sub.L	A0,D0		; get word length
	Bne	hq1		; if no data
	SubQ.L	#2,A6		; decriment buffer pointer
	Bra	hq2		; jump to end
HQ1:	Asr.L	#1,D0		; get word count
	Or.W	D0,(-2,A0)	; write word count
HQ2:
	.ENDM

	.macro Headset,N
	header = $8000 + N*32
	.endm

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;   Other Trivial Routines   ;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;



	.MACRO	Save_D0W  	 ; Write D0 to buffer (word)
	Move.W	D0, (A6)+	
	.ENDM


	.MACRO	Save_D0L	 ; Write D0 to buffer (long word)
	Swap    D0
	Move.L	D0, (A6)+	
	Swap	D0
	.ENDM



;;;;;;;;;;;;;;;;;;;;
;   Segmentation   ;
;;;;;;;;;;;;;;;;;;;;

	.MACRO	BEGREG,Regid	; Create Sub-segment header
	MoveA.L	A6,A1		; Store pointer
	Reghdr = Regid*256
	Move.W	#Reghdr,(A6)+	; Store Region Header
	.ENDM

	.MACRO	ENDREG		; Close Sub-segment
	Move.L	A6,D0		; Copy current pointer
	Sub.L	A1,D0		; Subtract start pointer
	Or.W	(A1),D0		; 
	Move.W	D0,(A1)		; Store Region Header
	.ENDM

.END
END_OF_FILE

if [ -e $ASM ]; then mv $ASM $ASM~; fi

$BINDIR/mac68030 $MAC 2>/dev/null <<'EOF'



EOF

$BINDIR/asm68030 $ASM 2>/dev/null <<'EOF'

EOF
