#include <stdio.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include "acc_conf.h"
#include "camac.h"
#include "k3922_reg.h"

#define BUFSIZE  32768/*in word*/
#define HDRSIZE (512/sizeof(short))
#define TXTSIZE (BUFSIZE-HDRSIZE)
#define DMA
union {
  struct {int base, trans;} info;
  struct {unsigned short header[HDRSIZE], text[TXTSIZE];} w;
} buf;

int main(int argc, char *argv[])
{
  int    fd, size, i;
  struct naf_data cc;

  setvbuf(stdout, (char *)NULL, _IONBF, 0); /* set stdout unbuffered */

  if (argc > 1)
    fd = open(argv[1], O_RDONLY);
  else {
    char file[256];
    printf("Executable image file > "); gets(file);
    fd = open(file, O_RDONLY);
  }
  if (fd == -1) {perror("image file"); exit(2);}
  size = read(fd, &buf, sizeof(buf));
  if (size <= 0) {perror("read image"); exit(2);}
  close(fd);
  printf("    %d bytes has been read.\n", size);
  printf("    Task base     address = 0x%6.6X\n", buf.info.base);
  printf("    Task transfer address = 0x%6.6X\n", buf.info.trans);

  if ((fd = open("/dev/cc0", O_RDWR)) == -1)
    {perror("CAMAC device"); exit(2);}

  ioctl(fd, CCIOC_SET_CRATE, ACC_CRATE);
  /* check crate on-line */
  cc.mode = SINGLE | WSIZE24;
  cc.naf  = NAFGEN(K3922_N,K3922_STAT_A,K3922_READ_F);
  ioctl(fd, CCIOC_SINGLE, &cc);
  if (cc.data & K3922_STAT_OFL) {puts("Crate off-line"); close(fd); exit(2);}
  /* clear inhibit (and set SRQ for the sake of BUGGY 2915) */
  cc.data = ~K3922_STAT_I & K3922_STAT_SRQ;
  cc.naf  = NAFGEN(K3922_N,K3922_STAT_A,K3922_WRIT_F);
  ioctl(fd, CCIOC_SINGLE, &cc);
  /* reset K3976 ACC */
  printf("    K3976 reset ... ");
  cc.mode = SINGLE | WSIZE16;
  cc.naf  = NAFGEN(ACC_N,0,9);
  ioctl(fd, CCIOC_SINGLE, &cc);
  /* wait for self-test done */
  printf("self-test ... ");
  cc.naf  = NAFGEN(ACC_N,15,8);
  do {
    usleep(100000L);
    ioctl(fd, CCIOC_SINGLE, &cc);
  } while (cc.xq & NOQ);
  puts("done.");

  /* set K3976 memory address counter (MAR) */
  cc.mode = SINGLE | WSIZE24;
  cc.data = buf.info.base;
  cc.naf  = NAFGEN(ACC_N,0,17);
  do ioctl(fd, CCIOC_SINGLE, &cc); while (cc.xq & (NOQ | NOX));
  printf("    Sending data to K3976 ACC ... ");
  size -= sizeof(buf.w.header);
  cc.naf  = NAFGEN(ACC_N,0,16);
#ifdef DMA
  cc.mode = QREPEAT | WSIZE16;
  ioctl(fd, CCIOC_SETUP_DMA, &cc);
  if ((i = write(fd, &(buf.w.text), size)) < 0)
    {perror("Down load"); close(fd); exit(2);}
#else
  cc.mode = SINGLE | WSIZE16;
  for (i = 0; i < size/sizeof(short); i++) {
    cc.data = buf.w.text[i];
    ioctl(fd, CCIOC_SINGLE, &cc);
  }
#endif
  puts("done.");

  /* set start address */
  cc.mode = SINGLE | WSIZE24;
  cc.data = buf.info.trans;
  cc.naf  = NAFGEN(ACC_N,0,17);
  ioctl(fd, CCIOC_SINGLE, &cc);

  /* start K3976 ACC */
  cc.naf  = NAFGEN(ACC_N,0,25);
  ioctl(fd, CCIOC_SINGLE, &cc);

  close(fd);
  exit(0);
}
