ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c								    c
c								    c
c	File : MAC68030.FOR					    c
c								    c
c   File converter from 68K macro file to 68K assembler file        c
c								    c
c	Target CPU: MOTROLA MC68030				    c
c	This is a cross compiler for MC68030 on VAX		    c
c	It is paired with ASM68030.FOR				    c
c								    c
c	Input file  : ****.MAC (Macro Assembler Code)		    c
c	Output file : ****.ASM (assembler like intermediate code)   c
c		 output file is in ascii character		    c
c		 so that you can check if you like		    c
c								    c
c				Author M.Iwasaki @ UTMSL	    c
c								    c
c				Version 1.1.01			    c
	character*7 		Version /'1.1.01 '/
c								    c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c									
c ##### SUPPORTED INSTRUCTIONS #####
c
c Generic Instructions
c
c 	Generic_Name	Mnemonic	Description
c   ------------------------------------------------------------
c	MOVE		MOVE		Move
c			MOVE CCR	Move CCR Register
c			MOVE SR		Move SR Register
c			MOVE USP	Move USP Register
c			MOVEA		Move to Address
c			MOVEQ		Move Quick
c			MOVEC		Move Control Register
c			PMOVE		Move MMU Register
c	MOVES		MOVES		Move Beyond super/user FC
c	AND		AND		Logical AND
c			ANDI		AND Immediate
c			ANDI CCR	AND Immediate to CCR Register
c			ANDI SR		AND Immediate to SR Register
c	OR		OR		Logical OR
c			ORI		OR Immediate
c			ORI CCR		OR Immediate to CCR Register
c			ORI SR		OR Immediate to SR Register
c	EOR		EOR		Exclusive OR
c			EORI		Exclusive OR Immediate
c			EORI CCR	Exclusive OR Immediate to CCR 
c			EORI SR		Exclusive OR Immediate to SR 
c	TST		TST		Test Operand
c	CMP		CMP		Compare
c			CMPA		Compare Address
c			CMPM		Compare Memory
c			CMPI		Compare Immediate
c	CLR		CLR		Clear
c	NOT		NOT		Complement
c	NEG		NEG		Negate
c			NEGX		Negate with Extend
c	ASR/L		ASR/L		Arithmetic Bit Shift
c	LSR/L		LSR/L		Logical Bit Shift
c	ROR/L		ROR/L		Bit Rotate
c	ROXR/L		ROXR/L		Bit Rotate with Extend
c	BTST		BTST		Bit Test
c	BCHG		BCHG		Bit Test & Change
c	BSET		BSET		Bit Test & Set
c	BCLR		BCLR		Bit Test & Clear
c	NOP		NOP		No Operation (wait Bus cycle)
c	CAS		CAS		Compare And Swap with operand
c	CAS2		CAS2		Compare And Swap with operand
c	ADD		ADD		Add
c			ADDA		Add to Address
c			ADDQ		Add Quick
c			ADDI		Add Immediate
c			ADDX		Add with Extend
c	SUB		SUB		Subtract
c			SUBA		Subtract from Address
c			SUBQ		Subtract Quick
c			SUBI		Subtract Immediate
c			SUBX		Subtract with Extend
c	SWAP		SWAP		Swap Register Ward
c	RESET		REST		Reset 
c
c Flow Control Instructions
c
c 	Generic_Name	Mnemonic	Description
c   ------------------------------------------------------------
c	BRA		BRA		Branch 
c	BSR		BSR		Branch to Subroutine
c	Bcc		Bcc		Branch on Condition )
c	DBcc		DBcc		Test, Decrement and Branch
c						(cc:=EQ,NE,... )
c	JMP		JMP		Jamp to Effective Address
c	JSR		JSR		Jamp to Subroutine
c	STOP		STOP		Stop
c
c MMU Instructions
c
c 	Generic_Name	Mnemonic	Description
c   ------------------------------------------------------------
c	PMOVE		PMOVE		Move MMU Register
c	PMOVEFD		PMOVEFD		Move MMU with Flush Disable
c	PFLUSH		PFLUSH		Flush ATC Entry
c	PFLUSHA		PFLUSHA		Flush All ATC Entry
c	PLOADR/W	PLOADR/W	Load ATC Entry as R/W Access
c	PTESTR/W	PTESTR/W	Logical Address Test
c	
c *** Instruction mode ***
c	instruction type     Mnemonic	op_mode(internal)
c	byte      		.B	  1
c	word      		.W	  2
c	long-word 		.L	  3 (default)
c	quad-word		.Q	  4
c
c
c
c ##### INTERNAL MODE DEFINITIONS #####
c
c *** Instruction mode ***
c	instruction type     Mnemonic	op_mode(internal)
c	byte      		.B	  1
c	word      		.W	  2
c	long-word 		.L	  3 (default)
c	quad-word		.Q	  4
c
c *** Data/Addressing Registers and its addressing mode *** ( SP = A7 )
c -----------------------------------------------------------------
c	addr_mode mode	register	addr_mode mode	register
c -----------------------------------------------------------------
c	Dn	  0	reg. No.	(xxx).W   7	0
c	An	  1	reg. No.	(xxx)[.L] 7	1
c	(An)	  2	reg. No.	#<data>   7	4
c	(An)+	  3	reg. No.
c	-(An)	  4	reg. No.
c	(d16,An)  5	reg. No.        (d16,PC)  7	2
c Note: (xxx) and (xxx).L is same (".L" is default)
c
c *** Control Registers (register number is internal) ***
c ------------------------------------------------------------------
c 	68K_reg   mode   register       68K_reg   mode   register
c ------------------------------------------------------------------
c	CCR	   -1	   10		CRP	   -1      17
c	SR	   -1      11		SRP	   -1      18
c	VBR	   -1      12		TC	   -1      19
c	SFC	   -1      13		TT0	   -1      20
c	DFC	   -1      14		TT1	   -1      21
c	CACR	   -1      15		MMUSR	   -1      22
c	CAAR	   -1      16		USP	   -1	   23
c				        MSP	   -1	   24
c				        ISP	   -1	   25
c ------------------------------------------------------------------
c mode : -1, register: -1   =  label direct (for branch instruction)
c mode : -2, register: -1   =  no argument or mismatch
c									
c
c
c ##### COMMAND (TITLE/CONTROL/DATA) #####
c
c Several control instructions are available to support programing.
c
c       format					description
c  -------------------------------------------------------------
c	.TITLE	title				title line
c						
c	.IF cnd_key, expression		    	If Block start
c	.ENDIF					If Block end
c
c	.REPEAT expression 			Repeat Block start
c	.ENDR					Repeat Block end
c
c       .MACRO  macro_name[, arg1, arg2, ...]	Macro Block start
c	.ENDM					Macro Block end
c
c	.WORD  value1[, value2, value3,...] 	word data line
c	.LWOAD value1[, value2, value3,...]	long-word data line
c
c	.END					terminator
c
c  *** .TITLE and .END lines ***
c TITLE line can contain a program title. To enable to put a in-line 
c comments, all the character after ";" mark will be ignored. 
c All the lines followed by .END line will be ignored.
c
c
c  *** Base Block ***
c Statements all lines except Macro Block is called as Base Block
c
c
c  *** If Block ***
c Statements in a "If Block" will be conditionally expanded in the 
c macro expansion. A "If Block" is started with .IF statement line and 
c terminated with .ENDIF statement line. The expression can be a 
c char & value type symbol or their complex.
c  
c	cnd_key      description
c    ------------------------------------------
c	  EQ	  equal to zero
c	  NE	  not equal to zero
c	  LT	  less than zero
c	  GT	  greater than zero
c	  LE	  less than or equal to zero
c	  GE	  greater than or equal to zero
c
c  *** Repeat Block ***
c Statements in a "Repeat Block" will be repeated as many time as 
c requested in the expression. If the expression is equal to or less 
c than zero, the Repeat Block will not be expanded in macro expansion. 
c A "Repeat Block" is started with .REPEAT statement line and 
c terminated with .ENDR statement line. The expression can be a 
c char & value type symbol or their complex. Note that the label can
c not be in tthe repeat block.
c
c  *** Macro Block ***
c Macro Block enables to name procedure (a set of machine codes) and 
c use that name like a machine code. A Macro Block has its name and 
c could have a numbers of (up to 28) arguments. Macro name can be up 
c to 21 characters.In this block .IF and .Repeat Block structure and 
c other MACRO call are also available. Although, recursive call is not 
c supported. The Macro Block will be expanded at each macro call when 
c the expansion. The arguments are evaluated on each macro call. 
c   A "Macro Block" is started with .MACRO statement line and 
c terminated with .ENDM statement line. Label can not be on .MACRO 
c statement line. Because it is ambiguous whether you intended it 
c is inside or outside of the Macro Block. The number of argument 
c should match between definition and its call.
c
c  *** DATA line Block ***
c DATA lines are available to put data directly. Number of values are 
c to up 29 per line. To avoid parity error in the instruction, byte 
c data definition is not supported,
c
c
c
c ##### SYMBOL DEFINITIONS #####
c
c Three symbol types are available namely Label, Value_symbol, 
c Macro_name and Macro_argument. All symbols must start with 
c alphabetic characters (A-Z). None of the symbol name can be same 
c as the instruction code nor the register name of 68030. 
c
c 1) Label symbol : for branch reference
c    symbol_type   type    format 	    limit of label name
c  -------------------------------------------------------------------
c    local_label   char    label_name:   6 characters (excluding ":")
c    global_label  char    label_name::  9 characters (excluding "::")
c
c	Note:   Local label is renamed with suffix numbered on macro 
c		expansion to avoid label duplication.
c	        LABEL is not allowed in MACRO definition line.
c		LABEL must be terminated with : or :: mark.
c
c 2) Value symbol (symbol in short)
c    symbol_type       type 	   format		
c  ---------------------------------------------------------------
c    value_symbol   char & value   symbol_name = value / expression
c
c      Note:    @ is a special value_symbol which defines the 
c		program address of the definition line.
c
c  *** available operator in the expression ***
c	Mnemonic	Description		Precedence
c     -----------------------------------------------------
c	   ^		power			first
c	   *		multiply		second
c	   /		divide			second
c	   +		add			third
c	   -		subtract or negative	third
c	   &		and			4th
c	   |		or			5th
c
c      Note: 	Symbols are evaluated on its definition. Symbols in 
c		the symbol defining expression must be pre-defined 
c		before its reference. The second definition of the 
c		same symbol name resulting symbol value replacement.
c
c 3) Macro symbol
c    symbol_type        type        format 
c  -------------------------------------------------------------------
c    macro_name      char           .MACRO  macro_name[, arg1, ...]
c    macro_argument  char & value   argument of .MACRO  
c
c      Note: 	Macro_arguments are substituted on the macro call.
c		If a macro_argument name is not defined as a 
c		value_symbol, it will be stored in symbol buffer 
c		and then the value will be substituted with that 
c		of the argument symbol. Macro name can be up to 21 
c		Macro name can be up to 21 characters.
c
c
c
c ##### ADDRESS DEFINITION #####
c
c Address of a label or a machine code can be specified directly using 
c special symbol definition @ = xxxx. The address of the successive line 
c is defined as xxxx. All other address is dynamically assigned. This
c feature is to support define address of data and label explicitly.
c
c
c
c ##### NUMERICAL EXPRESSION #####
c 
c Three numerical expressions are supported.
c 
c	format	  description
c    ---------------------------
c    [%][-]xxxx   decimal number 
c      $[-]xxxx   hex number 
c
c
c
c ##### UPPER CASE LOWER CASE #####
c
c There is no difference in upper case and lower case. (ex. ADD and AdD
c is same. In output file ".ASM", all character used in the input file 
c will be converted into lower case.
c
c
c





c 		Main Program MAC68K
c  Read a macro file and expend the line according to the 
c user definition and output a object file. 
c Create object library file (OLB) and a mapping file (MAP) in ascii
c on demand.
c local vars
	character*40 in_file_name, out_file_name
	character*40 map_file_name, olb_file_name
	character*255 cmdline
	integer ifc, ofc, mfc, lfc,lencmd
	character*1 ans	
	logical map_request, eval_line_request, olb_request
	logical read_error, expand_error
c begin
	write(*,*) ' --- Start MAC68030 --- Version : ', version
	open (unit=15, name='macro.cnd', status='NEW')
	  write(15,*) ' --- Start MAC68K --- Version : ', version
	close (15)
	call initialize_buffer
	map_request = .false.
	eval_line_request = .true.
	olb_request = .false.
	call lib$get_foreign(cmdline,,lencmd,)
	if(lencmd .gt. 0) then
	  in_file_name=cmdline(1:lencmd)
	  ifc=lencmd
	else
          write(*,'(1H$,A)') 'Enter Input File Name  > '
          read(5,'(Q,A)') ifc, in_file_name
	endif
	call generate_file_name(in_file_name, ifc, out_file_name, ofc, 
     &		      map_file_name, mfc, olb_file_name, lfc)
        write(*,'(1H$,A)')   'Build OLB file ? y/[n] > '
        read(5,'(A)') ans
        if(ans.eq.'y' .or. ans.eq.'Y') then
          write(*,'(1H$,A)') 'Enter OLB File Name    > '
          read(5,'(Q,A)') lfc, olb_file_name
	  olb_request = .true.
          if(olb_file_name(lfc-3:lfc-3) .ne. '.') then
	    olb_file_name(lfc+1:lfc+4) = '.OLB'
	    lfc = lfc + 4
	  endif
	else
          write(*,'(1H$,A)')   'Use OLB File ?   y/[n] > '
          read(5,'(A)') ans
          if(ans.eq.'y' .or. ans.eq.'Y') then
            write(*,'(1H$,A)') 'Enter OLB File Name    > '
            read(5,'(Q,A)') lfc, olb_file_name
            if(olb_file_name(lfc-3:lfc-3) .ne. '.') then
	      olb_file_name(lfc+1:lfc+4) = '.OLB'
	      lfc = lfc + 4
	    endif
	    open (unit=11, name=olb_file_name(1:lfc), status='OLD', err=10)
	      call read_olb
	    close (11)
	  endif
	endif
        write(*,'(1H$,A)')     'Output Map ?     y/[n] > '
        read(5,'(A)') ans
        if(ans.eq.'y' .or. ans.eq.'Y') then
	  map_request = .true.
	endif
c
	open (unit=10, name=in_file_name(1:ifc), status='OLD', err=20)
	  call read_macro(read_error)
	close (10)
	if(map_request) then
	  open (unit=11, name=map_file_name(1:mfc), status='NEW')
	    call output_map
	  close (11)
	endif
	if(.not.read_error) then
	  if(olb_request) then
	    open (unit=11, name=olb_file_name(1:lfc), status='NEW')
	      call output_olb
	    close (11)
	    stop
	  endif
	  open (unit=12, name=out_file_name(1:ofc), status='NEW')
	    call expand( eval_line_request, expand_error )
	  close (12)
	  if(.not. expand_error) then
	    call output_percentage
	    open (unit=15, name='macro.cnd', status='NEW')
	      write(15,'(A)') 'OK'
	    close (15)
	  endif
	endif
	stop
10	type *, 'Error: object Library File could not be found'
	stop
20	type *, 'Error: Input File could not be found'
	end





	subroutine generate_file_name(in_file_name, ifc, 
     &			out_file_name, ofc, 
     &		      	map_file_name, mfc, olb_file_name, lfc)
	implicit none
	character*40 in_file_name, out_file_name
	character*40 map_file_name, olb_file_name
	integer ifc, ofc, mfc, lfc
	logical find
	find = .false.
	ofc = 0
	Do While (.not. find .and. ofc.lt.ifc)
	  ofc = ofc + 1
	  if(in_file_name(ofc:ofc) .ne. '.') then
	    out_file_name(ofc:ofc) = in_file_name(ofc:ofc)
	    map_file_name(ofc:ofc) = in_file_name(ofc:ofc)
	    olb_file_name(ofc:ofc) = in_file_name(ofc:ofc)
	  else
	    find = .true.
	  endif
	EndDo
	if(find) then
	  out_file_name(ofc+1:ofc+4) = '.ASM'
	  map_file_name(ofc+1:ofc+4) = '.MAP'
	  olb_file_name(ofc+1:ofc+4) = '.OLB'
	  ofc = ofc + 4	  
	  mfc = ofc
	  lfc = ofc
	else
	  out_file_name(ofc+1:ofc+4) = '.ASM'
	  map_file_name(ofc+1:ofc+4) = '.MAP'
	  olb_file_name(ofc+1:ofc+4) = '.OLB'
	  in_file_name(ofc+1:ofc+4) = '.MAC'
	  ofc = ofc + 4	  
	  mfc = ofc
	  lfc = ofc
	  ifc = ofc
	endif
	return
	end



c------- subroutines to read macro file ----------------------------c



	subroutine read_macro(error)
c ... read MACRO file into buffer memory 
c
	implicit none
c argument
	integer base_end, macro_end
	logical srror
c temporary line buffer
	integer MAX_ARG
	parameter (MAX_ARG = 30)
	character*21 argument(MAX_ARG)
	integer al(MAX_ARG), n_argument
c local vars
	character*81 line
	integer lc, cntl, line_No, j
	integer cntl2, n_argument2
	logical in_macro_block, error
c	line_No : line No. in original file (for error message)
c	cntl : 1 = label line
c	cntl : 2 = symbol definition line
c	cntl : 3 = both 
c begin	
	in_macro_block = .false.
	line_No = 0
	error = .false.
c ... 	read until base end
	Do While(1)
5	  read(10,end=10,fmt='(Q,A)') lc, line
	  line_No = line_No + 1
	  if(lc.eq.0) go to 5
c ...	  separate a line into arguments
	  call argument_separation(line, lc, line_No, error, 
     &			cntl, argument, al, n_argument)
	  if(n_argument .eq. 0) go to 5
c	do j = 1, n_argument
c	write(1,*) j, '"'//argument(j)(1:al(j))//'"'
c	enddo
c ...	  ignore TITLE line and ignore all lines after END statement
	  if(argument(1)(1:al(1)) .eq. '.title') go to 5
	  if(argument(1)(1:al(1)) .eq. '.end') go to 10
	  if(n_argument.ne.0) then
c ... Label line : add label to buffer
	    if(mod(cntl,2).eq.1) then
	      call define_label(argument(1), al(1), line_No, error)
	    endif
	    if(mod(cntl,2).eq.1 .and. n_argument.gt.1) then
c ... Label on Macro definition is not allowed
	      if(argument(2)(1:al(2)) .eq. '.macro') then
c 	... 	.MACRO definition line can not be labeled (it is ambiguous!) 
	        type *, 'Error: Labeled .MACRO line'
	        write(*,*) '                     at line No. ', line_No
	        error = .true.
	      endif
c ... Divide Label and Instruction into two lines
	      cntl2 = int(cntl/2) * 2
	      cntl = 1
	      n_argument2 = n_argument - 1
	      n_argument = 1
c ... Labeled instruction/command line in a Macro Block
	      if(in_macro_block) then
		call put_macro_line(cntl, line_No,
     &				n_argument,  argument(1), al(1))
		call put_macro_line(cntl2, line_No,
     &				n_argument2, argument(2), al(2))
c ... Labeled instruction/command line in the Base block
	      else
		call put_base_line(cntl, line_No,
     &				n_argument,  argument(1), al(1))
		call put_base_line(cntl2, line_No,
     &				n_argument2, argument(2), al(2))
	      endif
	    else 
c ... Macro definition
	      if(argument(1)(1:al(1)) .eq. '.macro') then
		  call define_macro(cntl, line_No, 
     &			n_argument, argument, al, 
     &			in_macro_block, error)
c ... Macro termination
	      else if(argument(1)(1:al(1)) .eq. '.endm') then
		call terminate_macro(in_macro_block, line_No, error)
	      else
c ... Macro Block
	        if(in_macro_block) then
		  call put_macro_line(cntl, line_No,
     &				n_argument, argument, al)
c ... Base Block
	        else
		  call put_base_line(cntl, line_No,
     &				n_argument, argument, al)
	        endif
	      endif
	    endif
	  endif
	EndDo
10 	continue
	if(in_macro_block) then
	  type *, 'Error: MACRO is not terminated'
	  write(*,*) '                     at line No. ', line_No
	  error = .true.
	endif
	return
	end






	subroutine define_macro(cntl, line_No, 
     &			n_argument, argument, al, 
     &			in_macro_block, error)
	implicit none
c arguments
	integer cntl, n_argument
	integer al(n_argument)
	character*21 argument(n_argument)
	logical in_macro_block, error
	integer line_No
c Macro Block Buffer : This is for MACRO routines. Block
	integer MAX_MACRO, MAX_MACRO_POINTER, MAX_MACRO_NAME
	parameter (MAX_MACRO = 10000)
	parameter (MAX_MACRO_NAME = 500)
	parameter (MAX_MACRO_POINTER = 5000)
	integer macro_end, macro_name_end
	character*21 macro_name_buffer(MAX_MACRO_NAME)
	integer      macro_name_length(MAX_MACRO_NAME)
	integer      macro_name_pointer(MAX_MACRO_NAME)
	character*21 macro_code_buffer(MAX_MACRO)
	integer      macro_code_length(MAX_MACRO)
	integer macro_code_pointer(MAX_MACRO_POINTER)
	integer macro_cntl_marker(MAX_MACRO_POINTER)
	integer macro_line_No(MAX_MACRO_POINTER)
	common /MACRO/ macro_end, macro_code_buffer, macro_code_length,
     &		macro_code_pointer, macro_cntl_marker,
     &		macro_name_end, macro_name_pointer, macro_name_buffer,
     &		macro_name_length, macro_line_No
c local vars
	integer macro_p, j
	logical register, symbol, label, existing_macro, instruction
c begin
c	if it is MACRO definition line
	if(in_macro_block) then
	  type *, 'Error: .MACRO is in another Macro Block '
	  write(*,*) '                     at line No. ', line_No
	  error = .true.
	  return
	endif
	in_macro_block = .true.
	call check_register(argument(2), al(2), register)
	if(register) then
	  type *, 'Error: Macro name is conflicting with 68K Register name'
	  type *, '	Macro name:', argument(2)(1:al(2))
	  write(*,*) '                     at line No. ', line_No
	  error = .true.
	  return
	endif
	call check_symbol(argument(2), al(2), symbol)
	if(symbol) then
	  type *, 'Error: Macro name is conflicting with Symbol name'
	  type *, '	Macro name:', argument(2)(1:al(2))
	  write(*,*) '                     at line No. ', line_No
	  error = .true.
	  return
	endif
	call check_label(argument(2), al(2), label)
	if(label) then
	  type *, 'Error: Macro name is conflicting with Label name'
	  type *, '	Macro name:', argument(2)(1:al(2))
	  write(*,*) '                     at line No. ', line_No
	  error = .true.
	  return
	endif
	call check_instruction(argument(2), al(2), instruction)
	if(instruction) then
	  type *, 'Error: Macro name is conflicting with Instruction'
	  type *, '	Macro name:', argument(2)(1:al(2))
	  write(*,*) '                     at line No. ', line_No
	  error = .true.
	endif
	call check_macro(argument(2), al(2), existing_macro)
	if(existing_macro) then
	  type *, 'Error: multiple declaration of Macro name'
	  type *, '	Macro name:', argument(2)(1:al(2))
	  write(*,*) '                     at line No. ', line_No
	  error = .true.
	  return
	endif
	macro_end = macro_end + 1
	if( macro_end .ne. 1 ) then
	  macro_p = macro_code_pointer(macro_end)
	else
	  macro_p = 1
	  macro_code_pointer(macro_end) = macro_p 
	endif

	if(macro_end .lt. MAX_MACRO_POINTER) then
	  macro_name_end = macro_name_end + 1
	  if(macro_name_end .lt. MAX_MACRO_NAME) then
	    if(macro_p + n_argument - 2 .le. MAX_MACRO) then
	      macro_line_No(macro_end) = line_No
	      macro_name_length(macro_name_end) = al(2)
	      macro_name_buffer(macro_name_end)(1:al(2)) =
     &				argument(2)(1:al(2))
	      macro_name_pointer(macro_name_end) = macro_end
	      if(n_argument.ge.3) then
	        Do j = 3, n_argument
	          macro_code_buffer(macro_p) = argument(j)
	          macro_code_length(macro_p) = al(j)
	          call define_symbol(argument(j), al(j), 
     &				line_No, error)
		  macro_p = macro_p + 1
	        EndDo
c	      else
c		do nothing
	      endif
	      macro_code_pointer(macro_end+1) = macro_p
	    else
	      type *,'Fatal-Error: exceeded MAX_MACRO'
	      stop
	    endif
	  else
	    type *,'Fatal-Error: exceeded MAX_MACRO_NAME'
	    stop
	  endif
	else
	  type *,'Fatal-Error: exceeded MAX_MACRO_POINTER'
	  stop
	endif
	return
	end





	subroutine put_macro_line(cntl, Line_No, 
     &			n_argument, argument, al)
	implicit none
c arguments
	integer cntl, n_argument
	integer al(n_argument)
	character*21 argument(n_argument)
	integer line_No
c Macro Block Buffer : This is for MACRO routines. Block
	integer MAX_MACRO, MAX_MACRO_POINTER, MAX_MACRO_NAME
	parameter (MAX_MACRO = 10000)
	parameter (MAX_MACRO_NAME = 500)
	parameter (MAX_MACRO_POINTER = 5000)
	integer macro_end, macro_name_end
	character*21 macro_name_buffer(MAX_MACRO_NAME)
	integer      macro_name_length(MAX_MACRO_NAME)
	integer      macro_name_pointer(MAX_MACRO_NAME)
	character*21 macro_code_buffer(MAX_MACRO)
	integer      macro_code_length(MAX_MACRO)
	integer macro_code_pointer(MAX_MACRO_POINTER)
	integer macro_cntl_marker(MAX_MACRO_POINTER)
	integer macro_line_No(MAX_MACRO_POINTER)
	common /MACRO/ macro_end, macro_code_buffer, macro_code_length,
     &		macro_code_pointer, macro_cntl_marker,
     &		macro_name_end, macro_name_pointer, macro_name_buffer,
     &		macro_name_length, macro_line_No
c local vars
	integer macro_p, j
c begin
	macro_end = macro_end + 1
	if( macro_end .ne. 1 ) then
	  macro_p = macro_code_pointer(macro_end)
	else
	  macro_p = 1
	  macro_code_pointer(macro_end) = macro_p
	endif
	if(macro_end .lt. MAX_MACRO_POINTER) then
	  if(macro_p + n_argument .le. MAX_MACRO) then
	    macro_line_No(macro_end) = line_No
	    macro_cntl_marker(macro_end) = cntl
	    Do j = 1, n_argument
	      macro_code_buffer(macro_p) = argument(j)
	      macro_code_length(macro_p) = al(j)
	      macro_p = macro_p + 1
	    EndDo
	    macro_code_pointer(macro_end+1) = macro_p
	  else
	    type *,'Fatal-Error: exceeded MAX_MACRO'
	    stop
 	  endif
	else
	  type *,'Fatal-Error: exceeded MAX_MACRO_POINTER'
	  stop
	endif
	return
	end





	subroutine terminate_macro(in_macro_block, line_No, error)
	implicit none
c arguments 
	integer line_No
	logical in_macro_block, error
c Macro Block Buffer : This is for MACRO routines. Block
	integer MAX_MACRO, MAX_MACRO_POINTER, MAX_MACRO_NAME
	parameter (MAX_MACRO = 10000)
	parameter (MAX_MACRO_NAME = 500)
	parameter (MAX_MACRO_POINTER = 5000)
	integer macro_end, macro_name_end
	character*21 macro_name_buffer(MAX_MACRO_NAME)
	integer      macro_name_length(MAX_MACRO_NAME)
	integer      macro_name_pointer(MAX_MACRO_NAME)
	character*21 macro_code_buffer(MAX_MACRO)
	integer      macro_code_length(MAX_MACRO)
	integer macro_code_pointer(MAX_MACRO_POINTER)
	integer macro_cntl_marker(MAX_MACRO_POINTER)
	integer macro_line_No(MAX_MACRO_POINTER)
	common /MACRO/ macro_end, macro_code_buffer, macro_code_length,
     &		macro_code_pointer, macro_cntl_marker,
     &		macro_name_end, macro_name_pointer, macro_name_buffer,
     &		macro_name_length, macro_line_No
c begin
c ...	  if it is MACRO termination line
	  if(.not.in_macro_block) then
	    type *, 'Error: MACRO is not properly nested '
	    write(*,*) '                     at line No. ', line_No
	    error = .true.
	  endif
	  macro_name_pointer(macro_name_end+1) = macro_end + 1
	  in_macro_block = .false.
	return
	end





	subroutine put_base_line(cntl, line_No,
     &			n_argument, argument, al)
	implicit none
c arguments
	integer line_No
	integer cntl, n_argument
	integer al(n_argument)
	character*21 argument(n_argument)
c Base Block Buffer  : This is for body of the main stream.
	integer MAX_BASE, MAX_BASE_POINTER
	parameter (MAX_BASE = 10000)
	parameter (MAX_BASE_POINTER = 2000)
	integer base_end
	character*21 base_code_buffer(MAX_BASE)
	integer      base_code_length(MAX_BASE)
	integer      base_line_No(MAX_BASE_POINTER)
	integer base_code_pointer(MAX_BASE_POINTER)
	integer base_cntl_marker(MAX_BASE_POINTER)
	common /BASE/ base_end, base_code_buffer, base_code_length,
     &		base_code_pointer, base_cntl_marker, base_line_No
c local vars
	integer j, base_p
c begin
	base_end = base_end + 1
	if( base_end .ne. 1 ) then
	  base_p = base_code_pointer(base_end) 
	else
	  base_p = 1
	  base_code_pointer(base_end) = base_p
	endif
	if(base_end .lt. MAX_BASE_POINTER) then
	  if(base_p + n_argument .le. MAX_BASE) then
	    base_line_No(base_end) = Line_No
	    base_cntl_marker(base_end) = cntl
	    Do j = 1, n_argument
	      base_code_buffer(base_p) = argument(j)
	      base_code_length(base_p) = al(j)
	      base_p = base_p + 1
	    EndDo
	    base_code_pointer(base_end+1) = base_p
	  else
	    type *,'Fatal-Error: exceeded MAX_BASE'
	    stop
	  endif
	else
	  type *,'Fatal-Error: exceeded MAX_BASE_POINTER'
	  stop
	endif
	return
	end





	subroutine define_label(argument, al, line_No, error)
	implicit none
c local
	character*21 argument
	integer al, line_No
	logical error
c Label buffer : This is for label buffer.
	integer MAX_LABEL
	parameter (MAX_LABEL = 500)
	character*21 label_name(MAX_LABEL)
	integer label_end
	integer label_length(MAX_LABEL), gl_flag(MAX_LABEL)
	common /LBL/ label_end, label_name, label_length, gl_flag
c local variables
	integer j, gl, chk
	logical register, symbol, instruction, macro
c begin
	chk = al - 1
	if(argument(chk:chk).eq.':') then
c ...     global name
	  gl = 1
	  al = al - 2
	  if(al .gt. 9) then
	    type *, 'Error: Global Label ', argument(1:al),' is too long'
	    type *, '       Global Label is up to 9 characters'
	    error = .true.
	    return
	  endif
	else
c ...     local name
	  gl = 0
	  al = al - 1
	  if(al .gt. 6) then
	    type *, 'Error: Local Label ', argument(1:al), ' is too long'
	    type *, '       Local Label is up to 6 characters'
	    error = .true.
	    return
	  endif
	endif
	call check_register(argument, al, register)
	if(register) then
	  type *, 'Error: Label name is conflicting with 68K Register name'
	  type *, '	Label name:', argument(1:al)
	  write(*,*) '                     at line No. ', line_No
	  error = .true.
	  return
	endif
	call check_symbol(argument, al, symbol)
	if(symbol) then
	  type *, 'Error: Label name is conflicting with Symbol name'
	  type *, '	Label name:', argument(1:al)
	  write(*,*) '                     at line No. ', line_No
	  error = .true.
	  return
	endif
	call check_instruction(argument, al, instruction)
	if(instruction) then
	  type *, 'Error: Label name is conflicting with Instruction'
	  type *, '	Label name:', argument(1:al)
	  write(*,*) '                     at line No. ', line_No
	  error = .true.
	  return
	endif
	call check_macro(argument, al, macro)
	if(macro) then
	  type *, 'Error: Label name is conflicting with Macro name'
	  type *, '	Label name:', argument(1:al)
	  write(*,*) '                     at line No. ', line_No
	  error = .true.
	  return
	endif
	Do j = 1, label_end
	  if(label_name(j)(1:label_length(j)) .eq. argument(1:al)) then
	    type *, 'Error: LABEL duplicated at line No.', line_No
	    type *, '              Label Name: ' // argument(1:al)
	    error = .true.
	    return
	  endif
	EndDo
	label_end = label_end + 1
	if(label_end .le. MAX_LABEL) then
	  label_length(label_end) = al
	  label_name(label_end)(1:al) = argument(1:al)
	  gl_flag(label_end) = gl
	else
	  type *, 'Error: SYMBOL exceeds MAX_SYMBOL'
	  error = .true.
	  stop
	endif
	return
	end





	subroutine check_label(argument, al, find)
	implicit none
c local
	character*21 argument
	integer al
	logical find
c Label buffer : This is for label buffer.
	integer MAX_LABEL
	parameter (MAX_LABEL = 500)
	character*21 label_name(MAX_LABEL)
	integer label_end
	integer label_length(MAX_LABEL), gl_flag(MAX_LABEL)
	common /LBL/ label_end, label_name, label_length, gl_flag
c local variables
	integer j, gl, chk
	logical register, symbol, instruction
c begin
	find = .false.
	Do j = 1, label_end
	  if(label_length(j) .eq. al) then
	    if(label_name(j)(1:label_length(j)) 
     &			.eq. argument(1:al)) then
	      find = .true.
	      return
	    endif
	  endif
	EndDo
	return
	end





	subroutine define_symbol(symbol, ls, line_No, error)
	implicit none
c arguments
	character*21 symbol
	integer ls, line_No
c define symbol
	integer MAX_SYMBOL
	parameter (MAX_SYMBOL = 500)
	character*21 symbol_name(MAX_SYMBOL)
	integer symbol_end
	integer symbol_length(MAX_SYMBOL), symbol_value(MAX_SYMBOL)
	common /SYMBL/ symbol_end, symbol_name, symbol_length, symbol_value
c local variables
	logical register, label, instruction, macro, error
	integer j
c begin	
	call check_register(symbol, ls, register)
	if(register) then
	  type *, 'Error: Symbol name is conflicting with a 68K register name'
	  type *, '	Symbol name:', symbol(1:ls)
	  write(*,*) '                     at line No. ', line_No
	  error = .true.
	  return
	endif
	call check_instruction(symbol, ls, instruction)
	if(instruction) then
	  type *, 'Error: Symbol name is conflicting with instruction'
	  type *, '	Symbol name:', symbol(1:ls)
	  write(*,*) '                     at line No. ', line_No
	  error = .true.
	  return
	endif
	call check_label(symbol, ls, label)
	if(label) then
	  type *, 'Error: Symbol name is conflicting with label name'
	  type *, '	Symbol name:', symbol(1:ls)
	  write(*,*) '                     at line No. ', line_No
	  error = .true.
	  return
	endif
	call check_macro(symbol, ls, macro)
	if(macro) then
	  type *, 'Error: Symbol name is conflicting with Macro name'
	  type *, '	Symbol name:', symbol(1:ls)
	  write(*,*) '                     at line No. ', line_No
	  error = .true.
	  return
	endif
	Do j = 1, symbol_end
	  if(symbol_length(j) .eq. ls) then
	    if(symbol_name(j)(1:symbol_length(j)) .eq. symbol(1:ls)) then
	      return
	    endif
	  endif
	enddo
c	add new symbol
	symbol_end = symbol_end + 1
	if(symbol_end .le. MAX_SYMBOL) then
	  symbol_length(symbol_end) = ls
	  symbol_value(symbol_end) = -1
	  symbol_name(symbol_end) = symbol
	else
	  type *, 'Error: Symbol exceeds MAX_SYMBOL'
	  error = .true.
	  stop
	endif
	return
	end





	subroutine check_symbol(symbol, l_symbol, find)
	implicit none
c arguments
	character*21 symbol
	integer l_symbol
c define symbol
	integer MAX_SYMBOL
	parameter (MAX_SYMBOL = 500)
	character*21 symbol_name(MAX_SYMBOL)
	integer symbol_end
	integer symbol_length(MAX_SYMBOL), symbol_value(MAX_SYMBOL)
	common /SYMBL/ symbol_end, symbol_name, symbol_length, symbol_value
c local variables
	logical register, label, instruction, find
	integer j
c begin	
	find = .false.
	Do j = 1, symbol_end
	  if(symbol_length(j) .eq. l_symbol) then
	    if(symbol_name(j)(1:symbol_length(j)) .eq. symbol(1:l_symbol)) then
	      find = .true.
	      return
	    endif
	  endif
	EndDo
	return
	end





	subroutine check_macro(symbol, l_symbol, find)
	implicit none
c arguments
	character*21 symbol
	integer l_symbol
c Macro Block Buffer : This is for MACRO routines. Block
	integer MAX_MACRO, MAX_MACRO_POINTER, MAX_MACRO_NAME
	parameter (MAX_MACRO = 10000)
	parameter (MAX_MACRO_NAME = 500)
	parameter (MAX_MACRO_POINTER = 5000)
	integer macro_end, macro_name_end
	character*21 macro_name_buffer(MAX_MACRO_NAME)
	integer      macro_name_length(MAX_MACRO_NAME)
	integer      macro_name_pointer(MAX_MACRO_NAME)
	character*21 macro_code_buffer(MAX_MACRO)
	integer      macro_code_length(MAX_MACRO)
	integer macro_code_pointer(MAX_MACRO_POINTER)
	integer macro_cntl_marker(MAX_MACRO_POINTER)
	integer macro_line_No(MAX_MACRO_POINTER)
	common /MACRO/ macro_end, macro_code_buffer, macro_code_length,
     &		macro_code_pointer, macro_cntl_marker,
     &		macro_name_end, macro_name_pointer, macro_name_buffer,
     &		macro_name_length, macro_line_No
c local variables
	logical register, label, instruction, find
	integer j
c begin	
	find = .false.
c	type *, ' check macro *', symbol(1:l_symbol),'*'
	Do j = 1, macro_name_end
c	  type *, j, ':' // macro_name_buffer(j)(1:macro_name_length(j))
	  if(macro_name_length(j) .eq. l_symbol) then
	    if(macro_name_buffer(j)(1:macro_name_length(j)) .eq. 
     &		symbol(1:l_symbol)) then
	      find = .true.
	      return
	    endif
	  endif
	EndDo
	return
	end





	subroutine argument_separation(line, lc, line_No, error, 
     &			cntl, argument, al, n_argument)
	implicit none
c argument
	integer MAX_ARG
	parameter (MAX_ARG = 30)
	character*81 line
	character*21 argument(MAX_ARG)
	integer al(MAX_ARG), n_argument, lc
	logical error
c local
	integer label
	character*1 UA/'A'/, UZ/'Z'/, SPACE/' '/, COMMA/','/
	character*1 COLON/':'/, S_COLON/';'/
	character*1 PLUS/'+'/, MINUS/'-'/, MULT/'*'/
	character*1 DEV/'/'/, EQUIV/'='/, HAT/'^'/
	character*1 LK/'('/, RK/')'/
	character*1 COR/'|'/, CAND /'&'/
	integer n_colon, line_No, cntl, eval
	logical in_argument, equation, use_comma
	integer line_p, arg_p 
	character*1 chr
	byte bchr
	equivalence (bchr, chr)
c	use_comma : when true "," is used as argument separation character
c begin
	equation = .false.
	line_p = 1
	arg_p = 0
	n_colon = 0
	label = 0
	n_argument = 0
	eval = 0
	in_argument = .false.
	use_comma = .true.
c	type *, 'input: ', line(1:lc) 
	Do While(line(line_p:line_p) .ne. S_COLON .and. line_p .le. lc)
	  if( (line(line_p:line_p).le.SPACE) .or. 
     &	      (line(line_p:line_p).eq.COMMA).and.use_comma ) then
	    if(in_argument) then
	      if(n_argument.eq.1) then
	        if(argument(1)(1:arg_p) .eq. '.if') then
	          equation = .true.
	        else if(argument(1)(1:arg_p) .eq. '.repeat') then
	          equation = .true.	    
	        endif
	      endif
	      in_argument = .false.
	      al(n_argument) = arg_p
	      arg_p = 0
	    endif
	  else
	    if(.not.in_argument) then
	      n_argument = n_argument + 1
	      if(n_argument.gt.MAX_ARG) then
	        type *,'Error: Too many arguments'
	        write(*,*) '                     at line No. ', line_No
	        error = .true.
		return
	      endif
	    endif
	    if(line(line_p:line_p).eq.EQUIV) then
	      if(equation) then
	          type *,'Error: Too many "="'
	          write(*,*) '                     at line No. ', line_No
	          error = .true.
		  return
	      endif
	      equation = .true. 
	      if(in_argument) then
	        al(n_argument) = arg_p
	        n_argument = n_argument + 1
	        if(n_argument.gt.MAX_ARG) then
	          type *,'Error: Too many arguments '
	          write(*,*) '                     at line No. ', line_No
	          error = .true.
		  return
	        endif
	      endif
	      in_argument = .false.
	      argument(n_argument)(1:1) = line(line_p:line_p)
	      al(n_argument) = 1
	      arg_p = 0
	      eval = 2
	      call define_symbol(argument(n_argument-1), 
     &			al(n_argument-1), line_No, error)
	    else if(equation .and. (line(line_p:line_p).eq.PLUS .or. 
     &			line(line_p:line_p).eq.MINUS .or. 
     &			line(line_p:line_p).eq.CAND .or. 
     &			line(line_p:line_p).eq.COR .or. 
     &			line(line_p:line_p).eq.HAT .or. 
     &			line(line_p:line_p).eq.MULT .or. 
     &			line(line_p:line_p).eq.DEV) ) then
	      if(in_argument) then
	        al(n_argument) = arg_p
	        n_argument = n_argument + 1
	        if(n_argument.gt.MAX_ARG) then
	          type *,'Error: Too many arguments'
	          write(*,*) '                     at line No. ', line_No
	          error = .true.
		  return
	        endif
	      endif
	      argument(n_argument)(1:1) = line(line_p:line_p)
	      al(n_argument) = 1
	      in_argument = .false.
	      arg_p = 0
	    else
	      in_argument = .true.
	      arg_p = arg_p + 1
	      if(arg_p.gt.21) then
	        type *,'Error: Too long argument'
	        write(*,*) '                     at line No. ', line_No
	        error = .true.
		return
	      endif 
	      if(n_argument.eq.1 .and. line(line_p:line_p).eq.COLON) then
	        label = 1
	        n_colon = n_colon + 1
	      endif
	      if(line(line_p:line_p).ge.UA .and. 
     &		 line(line_p:line_p).le.UZ) then
		chr = line(line_p:line_p)
		bchr = bchr + 32
	        argument(n_argument)(arg_p:arg_p)  = chr
	      else
	        argument(n_argument)(arg_p:arg_p) = line(line_p:line_p)
	      endif
	    endif
	  endif
	  line_p = line_p + 1
	  if( n_argument-label .ge. 2 ) then
	    Do while (line(line_p:line_p).le.SPACE .and. line_p.le.80) 
	      line_p = line_p + 1
	    EndDo
	  endif
	  if(line(line_p:line_p).eq.LK) use_comma = .false. 
	  if(line(line_p:line_p).eq.RK) use_comma = .true. 
c	  if(arg_p.gt.0) then
c	    type *, line_p-1, n_argument, in_argument, use_comma, 
c     &			' '//line(line_p-1:line_p-1), 
c     &			' '//argument(n_argument)(1:arg_p)//'!'
c	  else
c	    type *, line_p-1, n_argument, in_argument, use_comma, 
c     &			' '//line(line_p-1:line_p-1)
c	  endif
	End Do
	if( in_argument ) then
	  al(n_argument) = arg_p
	endif
	if(n_colon.ge.3) then
	  type *,'Error: Wrong label'
	  write(*,*) '                     at line No. ', line_No
	  error = .true.
	  return
	endif
	cntl = label + eval
	return
	end




	subroutine check_instruction(code, cl, find)
	character*2 buf2 /'or'/
	character*3 buf3(61) /
	1	'add', 'and', 'asl', 'asr', 'bcc',
	1	'bcs', 'beq', 'bge', 'bgt', 'bhi',
	2	'ble', 'bls', 'blt', 'bmi', 'bne',
	2	'bpl', 'bvc', 'bvs', 'bra', 'bsr',
	3	'cas', 'chk', 'clr', 'cmp', 'clr',
	3	'cmp', 'exg', 'ext', 'jmp', 'jsr',
	4	'lea', 'lsl', 'lsr', 'neg', 'nop',
	4	'not', 'ori', 'pea', 'rol', 'ror',
	5	'rtd', 'rte', 'rtr', 'rts', 'scc',
	5	'scs', 'seq', 'sge', 'sgt', 'shi',
	6	'sle', 'sls', 'slt', 'smi', 'sne',
	6	'spl', 'svc', 'svs', 'sub', 'tas',
	1	'tst'/
	character*4 buf4(54) /
	1	'abcd', 'adda', 'addi', 'addq', 'addx',
	1	'andi', 'dbcc', 'dbcs', 'dbeq', 'dbge',
	2	'dbgt', 'dbhi', 'dble', 'dbls', 'dblt',
	2	'dbmi', 'dbne', 'dbpl', 'dbvc', 'dbvs',
	3	'bchg', 'bclr', 'bkpt', 'bset', 'btst',
	3	'cas2', 'chk2', 'cmpa', 'cmpi', 'cmpm',
	4	'cmp2', 'divs', 'divu', 'eori', 'extb',
	4	'link', 'move', 'muls', 'mulu', 'nbcd',
	5	'negx', 'pack', 'roxl', 'roxr', 'sbcd',
	5	'stop', 'suba', 'subi', 'subq', 'subx',
	4	'swap', 'trap', 'unlk', 'unpk'/
	character*5 buf5(18) / 
	1	'bfchg', 'bfclr', 'bfffo', 'bfins', 'bfset',
	1	'bftst', 'divsl', 'divul', 'movea', 'movec',
	2	'movem', 'movep', 'moveq', 'moves', 'pload',
	8	'ptest', 'reset', 'trapv'/
	character*6 buf6(19) / 
	1	'bfexts', 'bfextu', 'pflush', 'pmover', 'pmovew',
	1	'trapcc', 'trapcs', 'trapeq', 'trapge', 'trapgt',
	2	'traphi', 'traple', 'trapls', 'traplt', 'trapmi',
	9	'trapne', 'trappl', 'trapvc', 'trapvs'/
	character*7 buf7(3)/'illegal', 'pflusha', 'pmovefd'/
	character*21 code
	integer cl, j
	logical find
c begin
	find = .false.
	if(cl.eq.2) then
	  if(code(1:cl).eq.buf2) find = .true.
	else if(cl.eq.3) then
	  if(code(1:1).le.'j') then
	    j = 1
	    Do while (.not.find .and. j.le.30)
	      if(code(1:cl).eq.buf3(j)) find = .true.
	      j = j + 1
	    EndDo
	  else
	    j = 31
	    Do while (.not.find .and. j.le.61)
	      if(code(1:cl).eq.buf3(j)) find = .true.
	      j = j + 1
	    EndDo
	  endif
	else if(cl.eq.4) then
	  if(code(1:1).le.'c') then
	    j = 1
	    Do while (.not.find .and. j.le.25)
	      if(code(1:cl).eq.buf4(j)) find = .true.
	      j = j + 1
	    EndDo
	  else
	    j = 26
	    Do while (.not.find .and. j.le.54)
	      if(code(1:cl).eq.buf4(j)) find = .true.
	      j = j + 1
	    EndDo
	  endif
	else if(cl.eq.5) then
	  Do j = 1, 18
	    if(code(1:cl).eq.buf5(j)) find = .true.
	  EndDo
	else if(cl.eq.6) then
	  Do j = 1, 19
	    if(code(1:cl).eq.buf6(j)) find = .true.
	  EndDo
	else if(cl.eq.7) then
	  Do j = 1, 3
	    if(code(1:cl).eq.buf7(j)) find = .true.
	  EndDo
	endif
	return
	end



c------- subroutines to expand macro file --------------------------c



	subroutine expand( eval_line_request, expand_error )
c ... expand macros
c Procedure:
c 1: Evaluate symbol values in symbol definition/re-evaluation line.
c 2: Replace macro call to the referenced macro lines..
c 3: Modify (put suffix) to the local labels.
c 4: Evaluate the condition line.  If the condition is not "True", 
c    then omit the following block (If Block). Note that if the IF 
c    block is omitted, all symbol definition/calculation will also 
c    be ignored in this block. 
c 5: Evaluate the repeat line. Repeat the block (Repeat Block),  
c    as many times as the result value if the evaluation result 
c    in positive. 
c 6: Substitute symbol to values appeared in the instruction argument.
c 7: Output base asm lines. 
c
	implicit none
c argument
	logical eval_line_request, expand_error
c local vars
	integer i, j, k
	integer bpp, narg, cntl_marker, value
	character*21 code_buffer(0:29)
	integer code_length(0:29), n_op
	logical valid, macro_call, EOF, error
c	bpp		: line pointer (which is also be a pointer to 
c			  the pointer of code buffer)
c	narg		: number of arguments
c	cntl_marker	: indicating line type 
c				0: normal line (instruction/control)
c				1: label_line
c				2: symbol_definition/re-evaluation line
c	value		: value of a symbol
c	valid	: T: line_enable (valid), F: skip_line
	character*50 ermsg
	logical data_line, cntl_line
c begin
	error = .false.
	expand_error = .false.
	valid = .true.
	bpp = 0
	EOF = .false.
	call get_next(bpp, narg, cntl_marker, 
     &			code_length, code_buffer, EOF) 
	Do while (.not.EOF)
c	type *, code_buffer(0)(1:code_length(0))//' '//
c     &     code_buffer(1)(1:code_length(1)) //' '//
c     &     code_buffer(2)(1:code_length(2)) 
c ... Command line
	  if(code_buffer(0)(1:1) .eq. '.') then
	    call evaluate_command(bpp, valid, data_line, cntl_line,
     &				  narg, code_buffer, code_length, error)
	    expand_error = expand_error .or. error
	  else
	    data_line = .false.
	    cntl_line = .false.
	  endif
	  if(valid) then
c ... Control Line
	    if(cntl_line) then
c	      no_output
c ... Data Line
	    else if(data_line) then
	      call write_asm(code_buffer, code_length, narg) 
c ... Symbol line
	    else if(cntl_marker.eq.2) then
	      if(code_buffer(1)(1:code_length(1)) .ne. '=') then
	        ermsg = 'Error: wrong Symbol definition'
	        call write_line(bpp, ermsg)
		expand_error = .true.
	      endif
	      n_op = narg - 1
	      call evaluate_value(n_op, code_length(2), code_buffer(2), 
     &				value, error)
	      expand_error = expand_error .or. error
	      if(error) then
	        ermsg = 'Error: wrong symbol expression'
	        call write_line(bpp, ermsg)
	      endif
	      call set_val(code_buffer(0), code_length(0), value, error)
	      expand_error = expand_error .or. error
	      if(code_buffer(0)(1:code_length(0)) .eq. '@' .or.
     &		 code_buffer(0)(1:code_length(0)) .eq. '<' .or.
     &		              eval_line_request) then
	        call write_eval_line(code_buffer(0), code_length(0), value)
	      endif
c ... Label line
	    else if(cntl_marker.eq.1) then
	      call write_label( code_buffer(0), code_length(0) )
c ... Instruction line
	    else
c	...   check if the instruction is a macro call
	      if( .not. macro_call(bpp, eval_line_request, 
     &			      narg, code_buffer, code_length, error) ) then
	        Do j = 1, narg
c	...       replace symbol to value except label symbol
c 	          type *,'expa: in ', code_length(j), '  *'//code_buffer(j)//'*'
	          call replace_label_symbol(bpp, 
     &		    		code_buffer(j), code_length(j), error)
	          expand_error = expand_error .or. error
c	          type *,'expa: out ', code_length(j), '  *'//code_buffer(j)//'*'
	        EndDo
	        call write_asm(code_buffer, code_length, narg) 
	      endif
	      expand_error = expand_error .or. error
	    endif
	  endif
	  call get_next(bpp, narg, cntl_marker, 
     &			code_length, code_buffer, EOF)
	EndDo
	return
	end





	subroutine get_next(bpp, narg, cntl_marker, 
     &			code_length, code_buffer, EOF)
	implicit none
c arguments
	logical EOF
	integer bpp, cntl_marker
	integer code_length(0:29), narg
	character*21 code_buffer(0:29)
c Base Block Buffer  : This is for body of the main stream.
	integer MAX_BASE, MAX_BASE_POINTER
	parameter (MAX_BASE = 10000)
	parameter (MAX_BASE_POINTER = 2000)
	integer base_end
	character*21 base_code_buffer(MAX_BASE)
	integer      base_code_length(MAX_BASE)
	integer      base_line_No(MAX_BASE_POINTER)
	integer base_code_pointer(MAX_BASE_POINTER)
	integer base_cntl_marker(MAX_BASE_POINTER)
	common /BASE/ base_end, base_code_buffer, base_code_length,
     &		base_code_pointer, base_cntl_marker, base_line_No
c Macro Block Buffer : This is for MACRO routines. Block
	integer MAX_MACRO, MAX_MACRO_POINTER, MAX_MACRO_NAME
	parameter (MAX_MACRO = 10000)
	parameter (MAX_MACRO_NAME = 500)
	parameter (MAX_MACRO_POINTER = 5000)
	integer macro_end, macro_name_end
	character*21 macro_name_buffer(MAX_MACRO_NAME)
	integer      macro_name_length(MAX_MACRO_NAME)
	integer      macro_name_pointer(MAX_MACRO_NAME)
	character*21 macro_code_buffer(MAX_MACRO)
	integer      macro_code_length(MAX_MACRO)
	integer macro_code_pointer(MAX_MACRO_POINTER)
	integer macro_cntl_marker(MAX_MACRO_POINTER)
	integer macro_line_No(MAX_MACRO_POINTER)
	common /MACRO/ macro_end, macro_code_buffer, macro_code_length,
     &		macro_code_pointer, macro_cntl_marker,
     &		macro_name_end, macro_name_pointer, macro_name_buffer,
     &		macro_name_length, macro_line_No
c Macro Depth Buffer
	integer macro_depth, macro_depth_buffer(3,0:30) 
	common /MDEPTH/ macro_depth, macro_depth_buffer
c 	macro_depth 	: this is for macro call nest. It is used to generate 
c			  the suffix for the local labels
c local vars
	integer i, j, base_p
	logical first /.true./
	logical label
c begin
	if(first) then
	  macro_depth = 0
	  macro_depth_buffer(3,0) = 0 
	  first = .false.
	endif
	EOF = .false.
10	bpp = bpp + 1
	if(macro_depth.gt.0) then
	   if(bpp.gt.macro_depth_buffer(2,macro_depth)) then
	      macro_depth = macro_depth - 1
c	      type *,'macro end --- m_depth back to ', macro_depth
c	      type *,' bpp back from', bpp, ' to ', macro_depth_buffer(1,macro_depth)
	      bpp = macro_depth_buffer(1,macro_depth)
	      go to 10
	   endif
	endif
	if(macro_depth.eq.0) then
	    if(bpp.gt.base_end) then
	      EOF = .true.
	      return
	    endif
	    base_p = base_code_pointer(bpp)
	    narg = base_code_pointer(bpp+1) - 1  - base_p
	    cntl_marker = base_cntl_marker(bpp)
	    Do i = 0, narg
	      j = base_p + i
	      code_length(i) = base_code_length(j)
	      code_buffer(i)(1:code_length(i)) = 
     &			base_code_buffer(j)(1:base_code_length(j))
	    EndDo
	else
	    base_p = macro_code_pointer(bpp)
	    narg = macro_code_pointer(bpp+1) - 1 - base_p
	    cntl_marker = macro_cntl_marker(bpp)
	    Do i = 0, narg
	      j = base_p + i
	      code_length(i) = macro_code_length(j)
	      code_buffer(i)(1:code_length(i)) = 
     &			macro_code_buffer(j)(1:macro_code_length(j))
	    EndDo
	endif
	return
	end






	logical function macro_call(bpp, eval_line_request, 
     &			narg, code_buffer, code_length, error)
	implicit none
c arguments
	integer code_length(0:29), narg
	character*21 code_buffer(0:29)
	integer bpp
	logical eval_line_request
	logical error
c Macro Block Buffer : This is for MACRO routines. Block
	integer MAX_MACRO, MAX_MACRO_POINTER, MAX_MACRO_NAME
	parameter (MAX_MACRO = 10000)
	parameter (MAX_MACRO_NAME = 500)
	parameter (MAX_MACRO_POINTER = 5000)
	integer macro_end, macro_name_end
	character*21 macro_name_buffer(MAX_MACRO_NAME)
	integer      macro_name_length(MAX_MACRO_NAME)
	integer      macro_name_pointer(MAX_MACRO_NAME)
	character*21 macro_code_buffer(MAX_MACRO)
	integer      macro_code_length(MAX_MACRO)
	integer macro_code_pointer(MAX_MACRO_POINTER)
	integer macro_cntl_marker(MAX_MACRO_POINTER)
	integer macro_line_No(MAX_MACRO_POINTER)
	common /MACRO/ macro_end, macro_code_buffer, macro_code_length,
     &		macro_code_pointer, macro_cntl_marker,
     &		macro_name_end, macro_name_pointer, macro_name_buffer,
     &		macro_name_length, macro_line_No
c Macro Depth Buffer
	integer macro_depth, macro_depth_buffer(3,0:30) 
	common /MDEPTH/ macro_depth, macro_depth_buffer
c local vars
	integer i, j, jm, nargr, s_bpp, base_p, val 
	character*21 symbol
	integer ls, macro_call_count /0/
	character*50 ermsg
c begin
	Do i = 1, macro_name_end
	  if( code_buffer(0)(1:code_length(0)) .eq.
     &		macro_name_buffer(i)(1:macro_name_length(i)) ) then
	    macro_depth_buffer(1,macro_depth) = bpp
	    bpp = macro_name_pointer(i)
c	    type *,'macro call *',code_buffer(0)(1:code_length(0)),'*'
c	    type *,' bpp', macro_depth_buffer(1,macro_depth), ' to ', bpp
	    macro_depth = macro_depth + 1
c	    type *,' macro depth to be', macro_depth
	    macro_call_count = macro_call_count + 1
	    macro_depth_buffer(2,macro_depth) = macro_name_pointer(i+1) - 1
	    macro_depth_buffer(3,macro_depth) = macro_call_count
	    base_p = macro_code_pointer(bpp)
	    nargr = macro_code_pointer(bpp+1) - base_p
	    if(narg.ne.nargr) then
	      macro_depth = macro_depth - 1
	      s_bpp = macro_depth_buffer(1,macro_depth)
	      ermsg = 'Error: Macro argument No. is not consistent'
	      call write_line(s_bpp, ermsg)
	      macro_depth = macro_depth + 1
	      ermsg = '       Argument of this Macro defined as:'
	      call write_line(bpp, ermsg)
	      error = .true.
	      macro_call = .true.
	      return
	    endif
	    jm = base_p
	    Do j = 1, narg
	      call get_val(code_buffer(j), code_length(j), val, error)
	      ls = macro_code_length(jm)
	      symbol(1:ls) = 
     &			macro_code_buffer(jm)(1:macro_code_length(jm))
	      call set_val(symbol, ls, val, error)
	      if( eval_line_request ) then
	        call write_eval_line(symbol, ls, val)
	      endif
	      if(error) then
	        ermsg = 'Error: Macro argument is not defined yet'
	        call write_line(s_bpp, ermsg)
	      endif
	      jm = jm + 1
	    EndDo
	    macro_call = .true.
	    return
          endif
	EndDo
	macro_call = .false.
	return
	end





	subroutine evaluate_command(bpp, valid, data_line, cntl_line,
     &				    narg, code_buffer, code_length, error)
	implicit none
c arguments
	integer bpp
	logical valid, data_line, cntl_line, error
	integer narg
	character*21 code_buffer(0:19)
	integer code_length(0:19)
c local vars
	integer cond_depth/0/, cond_depth_buffer(3,30)
	integer disable_depth	
c 	cond_depth   	: this is for .IF or .REPEAT statement to indicate 
c			  how much of those are nested. It is used to check
c			  a line is valid or not, and to check the consistency
c			  of the nest
c	disable_depth	: cond_depth level where the expansion has been 
c			  disabled
	character*6 cntl_command(7) 
     &		/'if    ','repeat','endif ',
     &		 'endr  ','word  ','lword ','ermsg '/
	integer	    ctl(7) / 2,  6,  5,  4,  4,  5,  5 /
	integer k, n_op, value, minc, sp
	character*50 ermsg
c begin
	error = .false.
	data_line = .false.
	cntl_line = .false.
c ...   search control command
	k = 1
	Do while ( cntl_command(k)(1:ctl(k)) .ne. 
     &			code_buffer(0)(2:min(ctl(k)+1,code_length(0))) )
c	  type *,'code ', code_buffer(0)(2:min(ctl(k)+1,code_length(0))), 
c     &		ctl(k)+1, code_length(0), cntl_command(k)(1:ctl(k))
	  k = k + 1
	  if(k.gt.7) then
	    ermsg = 'Error: No such ".keyword" command supported'
	    call write_line(bpp, ermsg)
	    error = .true.
	    return
	  endif
	EndDo
c .TYPE
	if(k .eq. 7) then
	  cntl_line = .true.
	  if(valid) then
	    sp = 7
	    ermsg(1:sp) = 'Error: '
	    Do k = 1, Narg
	      ermsg(sp+1:sp+code_length(k)+1) 
     &			= code_buffer(k)(1:code_length(k)) // ' '
	      sp = sp + code_length(k) + 1
	    EndDo
	    type *, 'Msg: ',ermsg
	    error = .true.
	  endif
	  return
c .WORD or .LWORD
	else if(k .ge. 5) then
	  if(valid) then
	    data_line = .true.
	  else
	    cntl_line = .true.
	  endif
	  return
c ...   .IF or .REPEAT
	else if(k.le.2) then
	  cntl_line = .true.
	  cond_depth = cond_depth + 1
	  cond_depth_buffer(1, cond_depth) = k
	  if(valid) then
c .IF
	    if(k.eq.1) then
	      n_op = narg - 1
c 	...   evaluate argument expression
	      call evaluate_value(n_op, code_length(2), code_buffer(2), 
     &				value, error)
c	integer j
c	type *,'if', value
c	Do j = 1, n_op
c	type *, '"'//code_buffer(j+1)(1:code_length(j+1))//'"'
c	EndDo
	      if(error .or. code_length(1).ne.2) then
	          ermsg = 'Error: wrong IF expression'
	          call write_line(bpp, ermsg)
	 	  error = .true.
		  return
	      endif
c 	...   check value, if not set valid = .false.
	      call check_cnd(bpp, valid, cond_depth, 
     &			disable_depth, code_buffer(1), code_length(1), 
     &			value, error)
c	      type *,'* if', disable_depth, cond_depth, valid
	      if(error) return
c .REPEAT
	    else if(k.eq.2) then
c ...         evaluate repeat-number (argument)
	      call evaluate_value(narg, code_length(1), code_buffer(1), 
     &				value, error)
	      if(error) then
	          ermsg = 'Error: failed to calculate the repeat number'
	          call write_line(bpp, ermsg)
		  return
	      endif
	      if(value.gt.0) then 
c ...             set tern number and pointer position to be repeated
		  cond_depth_buffer(2, cond_depth) = value
		  cond_depth_buffer(3, cond_depth) = bpp
	      else
c ...             if argument is negative following lines
		  cond_depth_buffer(2, cond_depth) = 0
		  valid = .false.
		  disable_depth = cond_depth 
	      endif
c	      type *,'* repeat', disable_depth, cond_depth, 
c     &		cond_depth_buffer(2, cond_depth), valid
	    endif
	  endif
c       .ENDIF or .ENDR
	else 
	  cntl_line = .true.
	  k = k - 2
c	...check consistency
	  if(cond_depth_buffer(1, cond_depth) .ne. k) then
	      ermsg = 'Error: Condition is not correctly nested'
	      call write_line(bpp, ermsg)
	      error = .true.
	      return
	  endif
c .ENDIF
	  if(k .eq. 1) then
	    if(valid) then
	      cond_depth = cond_depth - 1	
	    else
	      if(cond_depth .eq. disable_depth) then
	        valid = .true.
	      endif
	      cond_depth = cond_depth - 1	
	    endif
c	    type *,'* endif', disable_depth, cond_depth, valid
c .ENDR
	  else if(k .eq. 2) then 
	    if(valid) then
	      cond_depth_buffer(2, cond_depth) 
     &			= cond_depth_buffer(2, cond_depth) - 1
	      if(cond_depth_buffer(2, cond_depth) .ge. 1) then
c		cond_depth = cond_depth
	        bpp = cond_depth_buffer(3, cond_depth)
	      else if(cond_depth_buffer(2, cond_depth) .eq. 0) then
		cond_depth = cond_depth - 1
	      endif
	    else
	      if(cond_depth .eq. disable_depth) then
		valid = .true.
	      endif
	      cond_depth = cond_depth - 1
	    endif
c	    type *,'* endr', disable_depth, cond_depth, 
c     &		cond_depth_buffer(2, cond_depth), valid
	  endif
	endif
c	type *,'* **', disable_depth, cond_depth, valid
	return
	end





	subroutine write_asm(code_buffer, code_length, narg) 
	implicit none
c arguments
	integer code_length(0:29), narg
	character*21 code_buffer(0:29)
c local vars
	integer cs_p, ce_p, jf
	character*80 line
	byte b_TAB, b_SPACE
	character*1 TAB, SPACE
	equivalence (TAB, b_TAB)
	equivalence (SPACE, b_SPACE)
c begin
	b_TAB = 9
	b_SPACE = 32
	line(1:1) = TAB
	cs_p = 2
	Do jf = 0, narg
	  ce_p = cs_p + code_length(jf) - 1
	  line(cs_p:ce_p) = code_buffer(jf)(1:code_length(jf))
	  ce_p = ce_p + 1
	  line(ce_p:ce_p) = TAB
	  cs_p = ce_p + 1
	EndDo
	write(12,'(A)') line(1:ce_p) 
	return
	end





	subroutine write_eval_line(symbol, ls, value)
	implicit none
c arguments
	character*21 symbol
	integer ls, value
c local vars
	character*80 line
	character*10 ivalue
	character*10 hvalue
	integer hl
	logical error
c begin
	write(ivalue, '(i10)') value
	call write_hex(hvalue, hl, value, error)
	line = 	'*' // symbol(1:ls) // ' = ' // 
     &		'  ' // hvalue(1:hl) // ' (hex)' // ivalue // ' (dec)'
	write(12, '(A)') line
	return
	end




	subroutine write_label(label, ll)
c ... write label // if label is global 
c ... write label_+_depth_mark // if label is local
	implicit none
c arguments
	character*21 label
	integer ll
c Label buffer : This is for label buffer.
	integer MAX_LABEL
	parameter (MAX_LABEL = 500)
	character*21 label_name(MAX_LABEL)
	integer label_end
	integer label_length(MAX_LABEL), gl_flag(MAX_LABEL)
	common /LBL/ label_end, label_name, label_length, gl_flag
c Macro Depth Buffer
	integer macro_depth, macro_depth_buffer(3,0:30) 
	common /MDEPTH/ macro_depth, macro_depth_buffer
c local vars
	integer i
	character*9 l_number
	logical error
c begin
	i = 1
	Do while (label(1:ll) .ne. label_name(i)(1:label_length(i)))
	  i = i + 1
	  if(i.gt.label_end) then
	    type *, ' Error: Label was not defined : ', label(1:ll)
	    error = .true.
	    return
	  endif
	EndDo
c ...   local flag
	if(gl_flag(i).eq.0) then
	  if(macro_depth_buffer(3,macro_depth) .ge. 
     &			10**(7-label_length(i)) ) then
	    type *,' Can you make the local label "',
     &			label_name(i)(1:label_length(i)),'" shoeter?'
	    type *,' Suffix area exceeded', 10**(7-label_length(i))
	    stop
	  endif
	  write(l_number,'(i9.9)') macro_depth_buffer(3,macro_depth)
	  ll = 9
	  label(1:ll) = label_name(i)(1:label_length(i)) // '_' //
     &				l_number(label_length(i)+2:9)
	endif
	write(12, '(A)') label(1:ll)
	return
	end





	subroutine replace_label_symbol(bpp, argument, length, error)
c ... Search label and symbol. If local-label is found, then replace it with
c ... label with local number.  If symbol is found replace them to its value.
	implicit none
c arguments
	character*21 argument
	integer length, bpp
c Label buffer : This is for label buffer.
	integer MAX_LABEL
	parameter (MAX_LABEL = 500)
	character*21 label_name(MAX_LABEL)
	integer label_end
	integer label_length(MAX_LABEL), gl_flag(MAX_LABEL)
	common /LBL/ label_end, label_name, label_length, gl_flag
c Symbol Buffer This is for symbol buffer.
	integer MAX_SYMBOL
	parameter (MAX_SYMBOL = 500)
	character*21 symbol_name(MAX_SYMBOL)
	integer symbol_end
	integer symbol_length(MAX_SYMBOL), symbol_value(MAX_SYMBOL)
	common /SYMBL/ symbol_end, symbol_name, symbol_length, symbol_value
c Macro Depth Buffer
	integer macro_depth, macro_depth_buffer(3,0:30) 
	common /MDEPTH/ macro_depth, macro_depth_buffer
c local vars
	integer i, val, lt
	character*9 l_number, label
	logical searching, error, register
	character*21 check, symbol
	integer lc, case, ls, lbn
	character*50 ermsg
c begin
	error = .false.
c ...	check if there is symbol or label in the argument (68K dependent)
c 	Select Symbol Region using the typical 68K addressing mode.
c	type *,'rep: argument','*'//argument(1:length)//'*'
	if(argument(1:1) .eq. '#') then
	  lc = length - 1
	  check(1:lc) = argument(2:length)
	  case = 1
	else if(argument(1:1) .eq. '(') then
	  if(argument(length-1:length-1) .eq. '.') then
	    lc = length - 4
	    check(1:lc) = argument(2:length-3)
	    case = 2
	    go to 10
	  endif
	  if(length .gt. 3) then
	    if(argument(length-3:length-3) .eq. ',') then
	      lc = length - 5
	      check(1:lc) = argument(2:length-4)
	      case = 3
	      go to 10
	    endif
	  endif
	  if(length .gt. 10) then
	    if(argument(length-10:length-10) .eq. ',') then
	      lc = length - 11
	      check(1:lc) = argument(2:length-7)
	      case = 4
	      go to 10
	    endif
	  endif
	  if(argument(length:length) .eq. ')') then
c	    type *,'check*','*'//argument(1:length)//'*'
	    lc = length - 2
	    check(1:lc) = argument(2:length-1)
	    call check_register(check, lc, register)
	    if(.not.register) then
	      case = 5
	      go to 10
	    else
	      return
	    endif
	  else
	    return
	  endif
	else if(argument(1:1) .eq. '-') then
	  if(length-3.eq.2 .and.
     &		argument(2:2) .eq. '(' .and.
     &		argument(length:length) .eq. ')' ) then
	    lc = 2
	    check(1:lc) = argument(3:length-1)
	    call check_register(check, lc, register)
	    if(register) then
	      return
	    else
	      error = .true.
	      type *, 'Error: Strange -(**) mode detected'
	      return
	    endif
	  else
	    error = .true.
	    type *, 'Error: Strange -(**) mode detected'
	    return
	  endif
	else
	  call check_register(argument, length, register)
	  if(register) then
	    return
	  else
	    case = 0
	    lc = length
	    check(1:lc) = argument(1:length)
	  endif
	endif
c Check Label Buffer
10	i = 1
c	type *,'*'//check(1:lc)//'*'
	searching = .true.
c ...	check if the argument is in the label
c	type *,'check* testing whether this is a flag or not'
	Do while (searching .and. i.le.label_end)
	  if(lc .eq. label_length(i)) then
	    if(check(1:lc) .eq. label_name(i)(1:lc)) then
	      searching = .false.
	      lbn = i
	    endif
	  endif
	  i = i + 1
	EndDo
c Label
	if(.not.searching) then
c   Local Label	: local flag needs suffix to avoid conflict due to 
c		  the multiple call of the Macro
	  if(gl_flag(lbn).eq.0) then
c	    type *,'check* this is a local flag'
	    write(l_number,'(i9.9)') macro_depth_buffer(3,macro_depth)
	    label(1:9) = label_name(lbn)(1:label_length(lbn)) // '_' // 
     &				l_number(label_length(lbn)+2:9)
	    if(case.eq.0) then
	      argument(1:9) = label(1:9)
	      length = 9
	    else if(case.eq.1) then
	      argument(1:10) = '#' // label(1:9)
	      length = 10
	    else if(case.eq.2) then
	      argument(1:13) = '(' // label(1:9) // argument(length-2:length)
	      length = 13
	    else if(case.eq.3) then
	      argument(1:14) = '(' // label(1:9) // argument(length-3:length)
	      length = 14
	    else if(case.eq.4) then
	      argument(1:21) = '(' // label(1:9) // argument(length-10:length)
	      length = 21
	    else if(case.eq.5) then
	      argument(1:13) = '(' // label(1:9) // ').l'
	      length = 13
	    endif
c   Global Label
c	  else
c	    do nothing for Global Label
	  endif
c Check Symbol Buffer
	else
c	  type *,'check* testing whether this is a symbol or not'
c	  type *,'check*'//check(1:lc)//'*check'
c Value_Symbol
	  call get_val(check, lc, val, error)
	  if(.not.error) then
c   Get Value of the Symbol
	    call write_hex(symbol, ls, val, error)
	    if(error) then
	      type *, 'Error: Negative value (out of range) for:',check(1:lc)    
	      error = .true.
	      return
	    endif
	    if(case.eq.0) then
	      lt = ls
	      argument(1:lt) = symbol(1:ls)
	    else if(case.eq.1) then
	      lt = 1 + ls
	      argument(1:lt) = '#' // symbol(1:ls)
	    else if(case.eq.2) then
	      if(argument(length-1:length-1) .eq.'.') then
	        lt = 4 + ls
	        argument(1:lt) = '(' // symbol(1:ls) 
     &				// argument(length-2:length)
	      else
	        lt = 2 + ls
	        argument(1:lt) = '(' // symbol(1:ls) 
     &				// argument(length:length)
	      endif
	    else if(case.eq.3) then
	      lt = 5 + ls
	      argument(1:lt) = '(' // symbol(1:ls) 
     &				// argument(length-3:length)
	    else if(case.eq.4) then
	      lt = ls + 12
	      argument(1:lt) = '(' // symbol(1:ls) 
     &				// argument(length-10:length)
	    else if(case.eq.5) then
	      argument(1:12) = '(' // symbol(1:ls) // ').l'
	      lt = ls + 4
	    endif
	    length = lt
c	    type *,'check* this is a symbol of value', argument(1:length),' in hex'
c This is NOT Label nor Symbol 
	  else
	    ermsg = 'Error: Symbol "'//check(1:lc)//'" is not defined yet'
	    call write_line(bpp, ermsg)
	  endif
	endif
	return
	end





	subroutine check_register(argument, length, register)
	character*21 argument
	integer length
	logical register
c local vars
	character*1 c
	byte b
	equivalence (b, c)
	integer j 
	character*5 reg(18) /
     &		'pc   ', 'ccr  ', 'sr   ', 'sp   ',
     &		'vbr  ', 'sfc  ', 'dfc  ', 'cacr ', 'caar ', 
     &		'crp  ', 'srp  ',
     &		'tc   ', 'tt0  ', 'tt1  ', 'mmusr',
     &		'usp', 'msp', 'isp'
     &		/
	integer rl(18) /2, 3, 2, 2,
     &		3, 3, 3, 4, 4,
     &		3, 3,
     &		2, 3, 3, 5, 3, 3, 3/
c begin
	register = .false.
	if(length .gt. 5) return
	if(length .eq. 2) then
	  c = argument(2:2)
	  b = b - 48
	  if( (argument(1:1) .eq. 'a' .or.  argument(1:1) .eq. 'd') .and.
     &	        (b .ge. 0 .and. b .le. 7) ) then
	    register = .true.
	    return
	  endif
	endif
	Do j = 1, 18
	  if( length .eq. rl(j) ) then
	    if( argument(1:length) .eq. reg(j)(1:rl(j)) ) then
	      register = .true.
	      return
	    endif
	  endif
	EndDo
	return
	end





	subroutine check_cnd(bpp, valid, depth, 
     &			disable_depth, symbol, ls, value, error)
	implicit none
c arguments
	character*21 symbol
	integer depth, disable_depth, value, ls, bpp
	logical valid, error
c local vars
	character*50 ermsg
c begin
	
	if(ls.ne.2) then
	  ermsg = 'Error: wrong if condition mark'
	  call write_line(bpp, ermsg)
	  error = .true.
	  return
	endif
	if(symbol(1:2) .eq. 'eq') then
	  if(value .eq. 0) then
	  else
	    valid = .false.
	    disable_depth = depth 
	  endif
	  return
	else if(symbol(1:2) .eq. 'ne') then
	  if(value .ne. 0) then
	  else
	    valid = .false.
	    disable_depth = depth 
	  endif
	  return
	else if(symbol(1:2) .eq. 'lt') then
	  if(value .lt. 0) then
	  else
	    valid = .false.
	    disable_depth = depth 
	  endif
	  return
	else if(symbol(1:2) .eq. 'gt') then
	  if(value .gt. 0) then
	  else
	    valid = .false.
	    disable_depth = depth 
	  endif
	  return
	else if(symbol(1:2) .eq. 'le') then
	  if(value .le. 0) then
	  else
	    valid = .false.
	    disable_depth = depth 
	  endif
	  return
	else if(symbol(1:2) .eq. 'ge') then
	  if(value .ge. 0) then
	  else
	    valid = .false.
	    disable_depth = depth 
	  endif
	  return
	endif
	ermsg = 'Error: wrong if condition mark'
	call write_line(bpp, ermsg)
	error = .true.
	return
	end





	subroutine write_line(bpp, ermsg)
c ... write line which is not in the macro statement
	implicit none
c arguments
	integer bpp
c Macro Depth Buffer
	integer macro_depth, macro_depth_buffer(3,0:30) 
	common /MDEPTH/ macro_depth, macro_depth_buffer
c Macro Block Buffer : This is for MACRO routines. Block
	integer MAX_MACRO, MAX_MACRO_POINTER, MAX_MACRO_NAME
	parameter (MAX_MACRO = 10000)
	parameter (MAX_MACRO_NAME = 500)
	parameter (MAX_MACRO_POINTER = 5000)
	integer macro_end, macro_name_end
	character*21 macro_name_buffer(MAX_MACRO_NAME)
	integer      macro_name_length(MAX_MACRO_NAME)
	integer      macro_name_pointer(MAX_MACRO_NAME)
	character*21 macro_code_buffer(MAX_MACRO)
	integer      macro_code_length(MAX_MACRO)
	integer macro_code_pointer(MAX_MACRO_POINTER)
	integer macro_cntl_marker(MAX_MACRO_POINTER)
	integer macro_line_No(MAX_MACRO_POINTER)
	common /MACRO/ macro_end, macro_code_buffer, macro_code_length,
     &		macro_code_pointer, macro_cntl_marker,
     &		macro_name_end, macro_name_pointer, macro_name_buffer,
     &		macro_name_length, macro_line_No
c Base Block Buffer  : This is for body of the main stream.
	integer MAX_BASE, MAX_BASE_POINTER
	parameter (MAX_BASE = 10000)
	parameter (MAX_BASE_POINTER = 2000)
	integer base_end
	character*21 base_code_buffer(MAX_BASE)
	integer      base_code_length(MAX_BASE)
	integer      base_line_No(MAX_BASE_POINTER)
	integer base_code_pointer(MAX_BASE_POINTER)
	integer base_cntl_marker(MAX_BASE_POINTER)
	common /BASE/ base_end, base_code_buffer, base_code_length,
     &		base_code_pointer, base_cntl_marker, base_line_No
c local vars
	integer cs_p, ce_p, j
	character*80 line
	byte b_TAB, b_SPACE
	character*1 TAB, SPACE
	equivalence (TAB, b_TAB)
	equivalence (SPACE, b_SPACE)
	character*50 ermsg
	integer line_No
c begin
	b_TAB = 9
	b_SPACE = 32
	if(macro_depth .eq. 0) then
	  line_No = base_line_No(bpp)
	  if(base_cntl_marker(bpp).eq.0) then
	      line(1:1) = TAB
	      cs_p = 2
	  else
	      cs_p = 1
	  endif
	  Do j = base_code_pointer(bpp), base_code_pointer(bpp+1) - 1
	    ce_p = cs_p + base_code_length(j) - 1
	    line(cs_p:ce_p) = base_code_buffer(j)(1:base_code_length(j))
	    ce_p = ce_p + 1
	    if(base_cntl_marker(bpp).eq.2) then
	      line(ce_p:ce_p) = SPACE
	    else
	      line(ce_p:ce_p) = TAB
	    endif
	    cs_p = ce_p + 1
	  EndDo
	  write(*,'(A)') '  '//ermsg
	  write(*,'(A)') '  '//line(1:ce_p) 
	  write(*,*) '                     at line No. ', line_No
	else
	  line_No = macro_line_No(bpp)
	  if(macro_cntl_marker(bpp).eq.0) then
	      line(1:1) = TAB
	      cs_p = 2
	  else
	      cs_p = 1
	  endif
	  Do j = macro_code_pointer(bpp), macro_code_pointer(bpp+1) - 1
	    ce_p = cs_p + macro_code_length(j) - 1
	    line(cs_p:ce_p) = macro_code_buffer(j)(1:macro_code_length(j))
	    ce_p = ce_p + 1
	    if(macro_cntl_marker(bpp).eq.2) then
	      line(ce_p:ce_p) = SPACE
	    else
	      line(ce_p:ce_p) = TAB
	    endif
	    cs_p = ce_p + 1
	  EndDo
	  write(*,'(A)') '  '//ermsg
	  write(*,'(A)') '  '//line(1:ce_p) 
	  write(*,*) '                     at line No. ', line_No
	endif
	return
	end





	subroutine evaluate_value(n_op, bufl, bufc, value, error)
c ... evaluate line 
	implicit none
c arguments
	integer n_op, value
	integer bufl(n_op)
	character*21 bufc(n_op)
	logical error
c local var
	integer MAX_ARG
	parameter (MAX_ARG = 30)
	character*1 symbol(7)/'^', '*', '/', '+', '-', '&', '|'/
	integer j, k, val(MAX_ARG), opf(MAX_ARG), valid(MAX_ARG)
	integer fl, a, b
c begin
c	write(1,*) 'evaluate'
c	Do j = 1, n_op
c	write(1,*) '"'//bufc(j)(1:bufl(j))//'"'
c	EndDo
c	error = .false.
	Do j = 1, n_op
	  valid(j) = 1
	  opf(j) = 0
	  if(bufl(j).eq.1) then
	    Do k = 1, 7
	      if(bufc(j)(1:1) .eq. symbol(k)) then
		opf(j) = k
	      endif
	    EndDo
	  endif
	  if(opf(j) .eq. 0) then
	    call get_val(bufc(j), bufl(j), val(j), error)
	    if(error) return
	  endif
	EndDo
	if(n_op .eq. 1) then
	  if(opf(1) .eq. 0) then
	    value = val(1)
	  else
	    error = .true.
	  endif
	  return
	endif
c Power
	Do j = 1, n_op
	  if(opf(j).eq.1) then
	    fl = 1
	    call get_valid(b, j, a, fl, valid, n_op, error)
	    val(b) = val(b)**val(a)
	    valid(a) = 0
	  endif
	EndDo
c Mult
	Do j = 1, n_op
	  if(opf(j).eq.2) then
	    fl = 1
	    call get_valid(b, j, a, fl, valid, n_op, error)
	    val(b) = val(b) * val(a)
	    valid(a) = 0
	  endif
	EndDo
c Dev
	Do j = 1, n_op
	  if(opf(j).eq.3) then
	    fl = 1
	    call get_valid(b, j, a, fl, valid, n_op, error)
	    val(b) = val(b) / val(a)
	    valid(a) = 0
	  endif
	EndDo
c Sub
	Do j = 1, n_op
	  if(opf(j).eq.5) then
	    fl = 0
	    call get_valid(b, j, a, fl, valid, n_op, error)
	    val(a) = - val(a)
	    opf(j) = 4
	  endif
	EndDo
c Add
	Do j = 1, n_op
	  if(opf(j).eq.4) then
	    if(j .eq. 1) then
	    else
	      fl = 1
	      call get_valid(b, j, a, fl, valid, n_op, error)
	      val(b) = val(b) + val(a)
	      valid(a) = 0
	    endif
	  endif
	EndDo
c And
	Do j = 1, n_op
	  if(opf(j).eq.6) then
	    fl = 1
	    call get_valid(b, j, a, fl, valid, n_op, error)
	    val(b) = val(b) .and. val(a)
	    valid(a) = 0
	  endif
	EndDo
c or
	Do j = 1, n_op
	  if(opf(j).eq.7) then
	    fl = 1
	    call get_valid(b, j, a, fl, valid, n_op, error)
	    val(b) = val(b) .or. val(a)
	    valid(a) = 0
	  endif
	EndDo
	if(opf(1).eq.0) then
	  value = val(1)
	else
	  value = val(2)
	endif
	return
	end





	subroutine get_valid(b, j, a, f_level, valid, n_op, error)
	implicit none
	integer MAX_ARG
	parameter (MAX_ARG = 30)
	integer b, j, a, valid(MAX_ARG), n_op, f_level
	logical error, find
c begin
	if(f_level.eq.1) valid(j) = 0
	find = .false.
	if(j.gt.1) then
	  b = j - 1
	  Do While (.not.find .and. b.ge.1)
	    if(valid(b).eq.1) find = .true.
	    b = b - 2
	  EndDo
	  b = b + 2
	  if(.not. find) error = .true.
	else
	  if(f_level.eq.1) error = .true.
	endif
	find = .false.
	if(j.lt.n_op) then
	  a = j + 1
	  Do While (.not.find .and. a.le.n_op)
	    if(valid(a).eq.1) find = .true.
	    a = a + 2
	  EndDo
	  a = a - 2
	  if(.not. find) error = .true.
	else
	  error = .true.
	endif
	return
	end





	subroutine get_val(symbol, ls, val, error)
	implicit none
c arguments
	character*21 symbol
	integer ls, val
	logical error
c Symbol Buffer This is for symbol buffer.
	integer MAX_SYMBOL
	parameter (MAX_SYMBOL = 500)
	character*21 symbol_name(MAX_SYMBOL)
	integer symbol_end
	integer symbol_length(MAX_SYMBOL), symbol_value(MAX_SYMBOL)
	common /SYMBL/ symbol_end, symbol_name, symbol_length, symbol_value
c locals
	integer j
	logical register
	logical*4 hval
	integer*4 ival
	equivalence (hval, ival)
c begin
	error = .false.
	if(symbol(1:1) .eq. '$') then
	  read(symbol(2:ls),err=10,fmt='(z)') hval
	else if(symbol(1:1) .eq. '%') then
	  read(symbol(2:ls),err=10,fmt='(i)') ival
	else if(symbol(1:1).ge.'0' .and. symbol(1:1).le.'9'
     &		.or. symbol(1:1).eq.'-') then
	  read(symbol(1:ls),err=10,fmt='(i)') ival
	else
	  Do j = 1, symbol_end
	    if( symbol_length(j).eq.ls ) then
	      if(symbol_name(j)(1:ls).eq.symbol(1:ls) ) then
	        val = symbol_value(j)
	        return
	      endif
	    endif
	  EndDo
	  error = .true.
	endif
	val = ival
	return
10	error = .true.
	type *, 'Getv: can not read value',symbol
	return
	end





	subroutine read_hex(symbol, ls, val, error)
	implicit none
c arguments
	character*(*) symbol
	integer ls, val
	logical error
c locals
	character*21 buf
	byte b_buf(21)
	equivalence (buf(1:1), b_buf(1))
	integer j, bc
	logical negative
c begin
	buf(1:ls) = symbol(1:ls)
	if(buf(1:1) .eq. '$') then
	  bc = 2
	else
	  bc = 1
	endif
	val = 0
	if(buf(bc:bc) .eq. '-') then
	  negative = .true.
	  bc = bc + 1
	else
	  negative = .false.
	endif	
	Do j = bc, ls
	  if(b_buf(j).ge.48 .and. b_buf(j).le.57) then
	    val = val * 16 + (b_buf(j) - 48)
	  else if(b_buf(j).ge.97 .and. b_buf(j).le.102) then
	    val = val * 16 + (b_buf(j) - 97) + 10
	  else
	    error = .true.
	    return
	  endif
	EndDo
	if(negative) val = - val
	return
	end





	subroutine write_hex(symbol, ls, val, error)
	implicit none
c arguments
	character*(*) symbol
	integer ls, val
	logical error
c locals
	character*21 buf
	byte b_buf(21)
	equivalence (buf(1:1), b_buf(1))
	integer j, tval, vbuf(21), i, d
	logical negative
c begin
	if(val.lt.0) then
	  negative = .true.
	  val = - val
	else
	  negative = .false.
	endif
	tval = val
	j = 0
	Do While (tval .ne. 0)
	  j = j + 1
	  d = tval / 16
	  vbuf(j) = tval - d*16
	  tval = d
	EndDo
	if(j.ne.0) then
	  Do i = j, 1, -1
	    if(vbuf(i).le.9) then
	      b_buf(j-i+1) = vbuf(i) + 48
	    else 
	      b_buf(j-i+1) = vbuf(i) - 10 + 97
	    endif
	  EndDo
	  ls = j + 1
	  if(negative) then
	    ls = ls + 1
	    symbol(1:ls) = '$-' // buf(1:j)
	  else
	    symbol(1:ls) = '$' // buf(1:j)
	  endif
	else
	  ls = 2
	  symbol(1:2) = '$0'
	Endif
	return
	end




	subroutine set_val(symbol, ls, val, error)
	implicit none
c arguments
	character*21 symbol
	integer ls, val
	logical error
c Symbol Buffer This is for symbol buffer.
	integer MAX_SYMBOL
	parameter (MAX_SYMBOL = 500)
	character*21 symbol_name(MAX_SYMBOL)
	integer symbol_end
	integer symbol_length(MAX_SYMBOL), symbol_value(MAX_SYMBOL)
	common /SYMBL/ symbol_end, symbol_name, symbol_length, symbol_value
c locals
	integer j
c begin
	error = .false.
	Do j = 1, symbol_end
	  if( symbol_length(j).eq.ls .and. 
     &		symbol_name(j)(1:ls).eq.symbol(1:ls) ) then
	    symbol_value(j) = val 
	    return
	  endif
	EndDo
	error = .true.
	return
	end



c------- subroutines for MAP and OLB output --------------------------c



	subroutine output_map
	implicit none
c argument
	integer MAX_ARG
	parameter (MAX_ARG = 30)
c Base Block Buffer  : This is for body of the main stream.
	integer MAX_BASE, MAX_BASE_POINTER
	parameter (MAX_BASE = 10000)
	parameter (MAX_BASE_POINTER = 2000)
	integer base_end
	character*21 base_code_buffer(MAX_BASE)
	integer      base_code_length(MAX_BASE)
	integer base_line_No(MAX_BASE_POINTER)
	integer base_code_pointer(MAX_BASE_POINTER)
	integer base_cntl_marker(MAX_BASE_POINTER)
	common /BASE/ base_end, base_code_buffer, base_code_length,
     &		base_code_pointer, base_cntl_marker, base_line_No
c Macro Block Buffer : This is for MACRO routines. Block
	integer MAX_MACRO, MAX_MACRO_POINTER, MAX_MACRO_NAME
	parameter (MAX_MACRO = 10000)
	parameter (MAX_MACRO_NAME = 500)
	parameter (MAX_MACRO_POINTER = 5000)
	integer macro_end, macro_name_end
	character*21 macro_name_buffer(MAX_MACRO_NAME)
	integer      macro_name_length(MAX_MACRO_NAME)
	integer      macro_name_pointer(MAX_MACRO_NAME)
	character*21 macro_code_buffer(MAX_MACRO)
	integer      macro_code_length(MAX_MACRO)
	integer macro_code_pointer(MAX_MACRO_POINTER)
	integer macro_cntl_marker(MAX_MACRO_POINTER)
	integer macro_line_No(MAX_MACRO_POINTER)
	common /MACRO/ macro_end, macro_code_buffer, macro_code_length,
     &		macro_code_pointer, macro_cntl_marker,
     &		macro_name_end, macro_name_pointer, macro_name_buffer,
     &		macro_name_length, macro_line_No
c Symbol Buffer This is for symbol buffer.
	integer MAX_SYMBOL
	parameter (MAX_SYMBOL = 500)
	character*21 symbol_name(MAX_SYMBOL)
	integer symbol_end
	integer symbol_length(MAX_SYMBOL), symbol_value(MAX_SYMBOL)
	common /SYMBL/ symbol_end, symbol_name, symbol_length, symbol_value
c Label buffer : This is for label buffer.
	integer MAX_LABEL
	parameter (MAX_LABEL = 500)
	character*21 label_name(MAX_LABEL)
	integer label_end
	integer label_length(MAX_LABEL), gl_flag(MAX_LABEL)
	common /LBL/ label_end, label_name, label_length, gl_flag
c local vars
	character*6 GL_symbol(0:1) /'Local ', 'Global'/  
	character*81 line
	integer bpp, mpp, k, j, cs_p, ce_p, n_op
	byte b_TAB, b_SPACE
	character*1 TAB, SPACE
	equivalence (TAB, b_TAB)
	equivalence (SPACE, b_SPACE)
	logical fline
c begin
	b_TAB = 9
	b_SPACE = 32
c writing bases
	write(11,*) '***** Base Block *****'
	Do bpp = 1, base_end
	  if(mod(base_cntl_marker(bpp),2).eq.0) then
	    line(1:1) = TAB
	    cs_p = 2
	  else
	    cs_p = 1
	  endif
	  Do j = base_code_pointer(bpp), base_code_pointer(bpp+1) - 1
	    ce_p = cs_p + base_code_length(j) - 1
	    line(cs_p:ce_p) = base_code_buffer(j)(1:base_code_length(j))
	    ce_p = ce_p + 1
	    if(base_cntl_marker(bpp).eq.2) then
	      line(ce_p:ce_p) = SPACE
	    else
	      line(ce_p:ce_p) = TAB
	    endif
	    cs_p = ce_p + 1
	  EndDo
	  write(11,*) base_line_No(bpp), '   ' // line(1:ce_p) 
	EndDo
c writing macro routines
	write(11,*) '***** Macro Block *****'
	if( macro_name_end .ge. 0) then
	  write(11,*) ' Number of Macro Routines :', macro_name_end
	  Do k = 1, macro_name_end
	    write(11,'(A)') 
     &		' ###  Macro Name: '//
	1	 macro_name_buffer(k)(1:macro_name_length(k))
	    fline = .true.
	    Do mpp = macro_name_pointer(k), macro_name_pointer(k+1) - 1
	      n_op = macro_code_pointer(mpp+1) - macro_code_pointer(mpp)
	      if(n_op .eq. 0 .and. fline) then
	        ce_p = 21
	        line(1:21) = ' --- No arguments ---'
	      else
	        if(fline) then
	 	  line(1:6) = ' Arg: '
	          cs_p = 7
	        else if(mod(macro_cntl_marker(mpp),2).eq.0) then
	          line(1:1) = TAB
	          cs_p = 2
	        else
	          cs_p = 1
	        endif
	        Do j = macro_code_pointer(mpp), macro_code_pointer(mpp+1) - 1
	          ce_p = cs_p + macro_code_length(j) - 1
	          line(cs_p:ce_p) = macro_code_buffer(j)
	1				(1:macro_code_length(j))
	          ce_p = ce_p + 1
	          if(macro_cntl_marker(mpp).eq.2) then
	            line(ce_p:ce_p) = SPACE
	          else
	            line(ce_p:ce_p) = TAB
	          endif
	          cs_p = ce_p + 1
	        EndDo
	      endif
	      write(11,*) macro_line_No(mpp), '   ' // line(1:ce_p) 
	      fline = .false.
	    EndDo
	  EndDo
	else
	  write(11, '(A)') ' --- No Macro Definition ---'
	endif
c writing symbols
	write(11,*) ' ***** Symbol Block *****'
	write(11,*) ' Number of Symbols :', symbol_end
	Do j = 1, symbol_end
	  write(11,'(A)') '   ' // symbol_name(j)(1:symbol_length(j))
	Enddo
c writing labels
	write(11,*) '***** Label Block *****'
	write(11,*) ' Number of Labels :', label_end
	Do j = 1, label_end
	  write(11,'(A)') '   ' // label_name(j)(1:label_length(j)) // ' : ' //
     &		GL_Symbol(gl_flag(j))
	Enddo
	return
	end





	subroutine output_percentage
	implicit none
c argument
	integer MAX_ARG
	parameter (MAX_ARG = 30)
c Base Block Buffer  : This is for body of the main stream.
	integer MAX_BASE, MAX_BASE_POINTER
	parameter (MAX_BASE = 10000)
	parameter (MAX_BASE_POINTER = 2000)
	integer base_end
	character*21 base_code_buffer(MAX_BASE)
	integer      base_code_length(MAX_BASE)
	integer      base_line_No(MAX_BASE_POINTER)
	integer base_code_pointer(MAX_BASE_POINTER)
	integer base_cntl_marker(MAX_BASE_POINTER)
	common /BASE/ base_end, base_code_buffer, base_code_length,
     &		base_code_pointer, base_cntl_marker, base_line_No
c Macro Block Buffer : This is for MACRO routines. Block
	integer MAX_MACRO, MAX_MACRO_POINTER, MAX_MACRO_NAME
	parameter (MAX_MACRO = 10000)
	parameter (MAX_MACRO_NAME = 500)
	parameter (MAX_MACRO_POINTER = 5000)
	integer macro_end, macro_name_end
	character*21 macro_name_buffer(MAX_MACRO_NAME)
	integer      macro_name_length(MAX_MACRO_NAME)
	integer      macro_name_pointer(MAX_MACRO_NAME)
	character*21 macro_code_buffer(MAX_MACRO)
	integer      macro_code_length(MAX_MACRO)
	integer macro_code_pointer(MAX_MACRO_POINTER)
	integer macro_cntl_marker(MAX_MACRO_POINTER)
	integer macro_line_No(MAX_MACRO_POINTER)
	common /MACRO/ macro_end, macro_code_buffer, macro_code_length,
     &		macro_code_pointer, macro_cntl_marker,
     &		macro_name_end, macro_name_pointer, macro_name_buffer,
     &		macro_name_length, macro_line_No
c Symbol Buffer This is for symbol buffer.
	integer MAX_SYMBOL
	parameter (MAX_SYMBOL = 500)
	character*21 symbol_name(MAX_SYMBOL)
	integer symbol_end
	integer symbol_length(MAX_SYMBOL), symbol_value(MAX_SYMBOL)
	common /SYMBL/ symbol_end, symbol_name, symbol_length, symbol_value
c Label buffer : This is for label buffer.
	integer MAX_LABEL
	parameter (MAX_LABEL = 500)
	character*21 label_name(MAX_LABEL)
	integer label_end
	integer label_length(MAX_LABEL), gl_flag(MAX_LABEL)
	common /LBL/ label_end, label_name, label_length, gl_flag
c begin
	write(*,*) '**Resource**'
	write(*,'(A,i8,A,f8.2,A)') 
     &	' allocated MAX_BASE:         ',MAX_BASE,
     &	'  used :',100.*base_code_pointer(base_end+1)/MAX_BASE,' %'
	write(*,'(A,i8,A,f8.2,A)') 
     &	' allocated MAX_BASE_POINTER: ',MAX_BASE_POINTER,
     &	'  used :',100.*(base_end + 1)/MAX_BASE_POINTER,' %'
	write(*,'(A,i8,A,f8.2,A)') 
     &	' allocated MAX_MACRO:        ',MAX_MACRO,
     &	'  used :',100.*macro_code_pointer(macro_end+1)/MAX_MACRO,' %'
	write(*,'(A,i8,A,f8.2,A)') 
     &	' allocated MAX_MACRO_POINTER:',MAX_MACRO_POINTER,
     &	'  used :',100.*(macro_end + 1)/MAX_MACRO_POINTER,' %'
	write(*,'(A,i8,A,f8.2,A)') 
     &	' allocated MAX_MACRO_NAME:   ',MAX_MACRO_NAME,
     &	'  used :',100.*macro_name_end/MAX_MACRO_NAME,' %'
	write(*,'(A,i8,A,f8.2,A)') 
     &	' allocated MAX_SYMBOL:       ',MAX_SYMBOL,
     &	'  used :',100.*symbol_end/MAX_SYMBOL,' %'
	write(*,'(A,i8,A,f8.2,A)') 
     &	' allocated MAX_LABEL:        ',MAX_LABEL,
     &	'  used :',100.*label_end/MAX_LABEL,' %'
	return
	end





	subroutine output_olb
	implicit none
c Base Block Buffer  : This is for body of the main stream.
	integer MAX_BASE, MAX_BASE_POINTER
	parameter (MAX_BASE = 10000)
	parameter (MAX_BASE_POINTER = 2000)
	integer base_end
	character*21 base_code_buffer(MAX_BASE)
	integer      base_code_length(MAX_BASE)
	integer      base_line_No(MAX_BASE_POINTER)
	integer base_code_pointer(MAX_BASE_POINTER)
	integer base_cntl_marker(MAX_BASE_POINTER)
	common /BASE/ base_end, base_code_buffer, base_code_length,
     &		base_code_pointer, base_cntl_marker, base_line_No
c Macro Block Buffer : This is for MACRO routines. Block
	integer MAX_MACRO, MAX_MACRO_POINTER, MAX_MACRO_NAME
	parameter (MAX_MACRO = 10000)
	parameter (MAX_MACRO_NAME = 500)
	parameter (MAX_MACRO_POINTER = 5000)
	integer macro_end, macro_name_end
	character*21 macro_name_buffer(MAX_MACRO_NAME)
	integer      macro_name_length(MAX_MACRO_NAME)
	integer      macro_name_pointer(MAX_MACRO_NAME)
	character*21 macro_code_buffer(MAX_MACRO)
	integer      macro_code_length(MAX_MACRO)
	integer macro_code_pointer(MAX_MACRO_POINTER)
	integer macro_cntl_marker(MAX_MACRO_POINTER)
	integer macro_line_No(MAX_MACRO_POINTER)
	common /MACRO/ macro_end, macro_code_buffer, macro_code_length,
     &		macro_code_pointer, macro_cntl_marker,
     &		macro_name_end, macro_name_pointer, macro_name_buffer,
     &		macro_name_length, macro_line_No
c Symbol Buffer This is for symbol buffer.
	integer MAX_SYMBOL
	parameter (MAX_SYMBOL = 500)
	character*21 symbol_name(MAX_SYMBOL)
	integer symbol_end
	integer symbol_length(MAX_SYMBOL), symbol_value(MAX_SYMBOL)
	common /SYMBL/ symbol_end, symbol_name, symbol_length, symbol_value
c Label buffer : This is for label buffer.
	integer MAX_LABEL
	parameter (MAX_LABEL = 500)
	character*21 label_name(MAX_LABEL)
	integer label_end
	integer label_length(MAX_LABEL), gl_flag(MAX_LABEL)
	common /LBL/ label_end, label_name, label_length, gl_flag
c local vars
	character*81 line
	integer bpp, mpp, k, j, cs_p, ce_p, n_op
c begin
c writing bases
	write(11,'(i8)') base_end
	Do bpp = 1, base_end
	  write(11,'(i3, i8, i8)') base_cntl_marker(bpp), 
     &		base_code_pointer(bpp), base_code_pointer(bpp+1)
	  Do j = base_code_pointer(bpp), base_code_pointer(bpp+1) - 1
	    write(11,'(i3)') base_code_length(j)
	    write(11,'(A)') base_code_buffer(j)(1:base_code_length(j))
	  EndDo
	EndDo
c writing macro routines
	write(11,'(i8)') macro_end
	write(11,'(i8)') macro_name_end
	Do k = 1, macro_name_end
	  write(11,'(i3)') macro_name_length(k)
	  write(11,'(A)')  macro_name_buffer(k)(1:macro_name_length(k))
	  write(11,'(i8, i8)') 
     &		macro_name_pointer(k), macro_name_pointer(k+1)
	  Do mpp = macro_name_pointer(k), macro_name_pointer(k+1) - 1
	    write(11,'(i3, i8, i8)') macro_cntl_marker(mpp), 
     &		macro_code_pointer(mpp), macro_code_pointer(mpp+1)
	    Do j = macro_code_pointer(mpp), macro_code_pointer(mpp+1) - 1
	      write(11,'(i3)') macro_code_length(j)
	      write(11,'(A)') macro_code_buffer(j)(1:macro_code_length(j))
	    EndDo
	  EndDo
	EndDo
c writing symbols
	write(11,'(i8)') symbol_end
	Do j = 1, symbol_end
	  write(11,'(i3)') symbol_length(j)
	  write(11,'(A)') symbol_name(j)(1:symbol_length(j))
	Enddo
c writing labels
	write(11,'(i8)') label_end
	Do j = 1, label_end
	  write(11,'(2i3)') gl_flag(j), label_length(j)
	  write(11,'(A)') label_name(j)(1:label_length(j)) 
	Enddo
	return
	end





	subroutine read_olb
	implicit none
c Base Block Buffer  : This is for body of the main stream.
	integer MAX_BASE, MAX_BASE_POINTER
	parameter (MAX_BASE = 10000)
	parameter (MAX_BASE_POINTER = 2000)
	integer base_end
	character*21 base_code_buffer(MAX_BASE)
	integer      base_code_length(MAX_BASE)
	integer      base_line_No(MAX_BASE_POINTER)
	integer base_code_pointer(MAX_BASE_POINTER)
	integer base_cntl_marker(MAX_BASE_POINTER)
	common /BASE/ base_end, base_code_buffer, base_code_length,
     &		base_code_pointer, base_cntl_marker, base_line_No
c Macro Block Buffer : This is for MACRO routines. Block
	integer MAX_MACRO, MAX_MACRO_POINTER, MAX_MACRO_NAME
	parameter (MAX_MACRO = 10000)
	parameter (MAX_MACRO_NAME = 500)
	parameter (MAX_MACRO_POINTER = 5000)
	integer macro_end, macro_name_end
	character*21 macro_name_buffer(MAX_MACRO_NAME)
	integer      macro_name_length(MAX_MACRO_NAME)
	integer      macro_name_pointer(MAX_MACRO_NAME)
	character*21 macro_code_buffer(MAX_MACRO)
	integer      macro_code_length(MAX_MACRO)
	integer macro_code_pointer(MAX_MACRO_POINTER)
	integer macro_cntl_marker(MAX_MACRO_POINTER)
	integer macro_line_No(MAX_MACRO_POINTER)
	common /MACRO/ macro_end, macro_code_buffer, macro_code_length,
     &		macro_code_pointer, macro_cntl_marker,
     &		macro_name_end, macro_name_pointer, macro_name_buffer,
     &		macro_name_length, macro_line_No
c Symbol Buffer This is for symbol buffer.
	integer MAX_SYMBOL
	parameter (MAX_SYMBOL = 500)
	character*21 symbol_name(MAX_SYMBOL)
	integer symbol_end
	integer symbol_length(MAX_SYMBOL), symbol_value(MAX_SYMBOL)
	common /SYMBL/ symbol_end, symbol_name, symbol_length, symbol_value
c Label buffer : This is for label buffer.
	integer MAX_LABEL
	parameter (MAX_LABEL = 500)
	character*21 label_name(MAX_LABEL)
	integer label_end
	integer label_length(MAX_LABEL), gl_flag(MAX_LABEL)
	common /LBL/ label_end, label_name, label_length, gl_flag
c local vars
	character*81 line
	integer bpp, mpp, k, j, cs_p, ce_p, n_op
c begin
c writing bases
	read(11,'(i8)') base_end
	Do bpp = 1, base_end
	  read(11,'(i3, i8, i8)') base_cntl_marker(bpp), 
     &		base_code_pointer(bpp), base_code_pointer(bpp+1)
	  Do j = base_code_pointer(bpp), base_code_pointer(bpp+1) - 1
	    read(11,'(i3)') base_code_length(j) 
	    read(11,'(A)') base_code_buffer(j)(1:base_code_length(j))
	  EndDo
	EndDo
c writing macro routines
	read(11,'(i8)') macro_end
	read(11,'(i8)') macro_name_end
	Do k = 1, macro_name_end
	  read(11,'(i3)') macro_name_length(k)
	  read(11,'(A)') macro_name_buffer(k)(1:macro_name_length(k))
	  read(11,'(i8, i8)') 
     &		macro_name_pointer(k), macro_name_pointer(k+1)
	  Do mpp = macro_name_pointer(k), macro_name_pointer(k+1) - 1
	    read(11,'(i3, i8, i8)') macro_cntl_marker(mpp), 
     &		macro_code_pointer(mpp), macro_code_pointer(mpp+1)
	    Do j = macro_code_pointer(mpp), macro_code_pointer(mpp+1) - 1
	        read(11,'(i3)') macro_code_length(j)
	        read(11,'(A)') macro_code_buffer(j)(1:macro_code_length(j))
	    EndDo
	  EndDo
	EndDo
c writing symbols
	read(11,'(i8)') symbol_end
	Do j = 1, symbol_end
	  read(11,'(i3)') symbol_length(j)
	  read(11,'(A)') symbol_name(j)(1:symbol_length(j))
	Enddo
c writing labels
	read(11,'(i8)') label_end
	Do j = 1, label_end
	  read(11,'(2i3)') gl_flag(j), label_length(j)
	  read(11,'(A)') label_name(j)(1:label_length(j)) 
	Enddo
	return
	end





	subroutine initialize_buffer
	implicit none
c Base Block Buffer  : This is for body of the main stream.
	integer MAX_BASE, MAX_BASE_POINTER
	parameter (MAX_BASE = 10000)
	parameter (MAX_BASE_POINTER = 2000)
	integer base_end
	character*21 base_code_buffer(MAX_BASE)
	integer      base_code_length(MAX_BASE)
	integer      base_line_No(MAX_BASE_POINTER)
	integer base_code_pointer(MAX_BASE_POINTER)
	integer base_cntl_marker(MAX_BASE_POINTER)
	common /BASE/ base_end, base_code_buffer, base_code_length,
     &		base_code_pointer, base_cntl_marker, base_line_No
c Macro Block Buffer : This is for MACRO routines. Block
	integer MAX_MACRO, MAX_MACRO_POINTER, MAX_MACRO_NAME
	parameter (MAX_MACRO = 10000)
	parameter (MAX_MACRO_NAME = 500)
	parameter (MAX_MACRO_POINTER = 5000)
	integer macro_end, macro_name_end
	character*21 macro_name_buffer(MAX_MACRO_NAME)
	integer      macro_name_length(MAX_MACRO_NAME)
	integer      macro_name_pointer(MAX_MACRO_NAME)
	character*21 macro_code_buffer(MAX_MACRO)
	integer      macro_code_length(MAX_MACRO)
	integer macro_code_pointer(MAX_MACRO_POINTER)
	integer macro_cntl_marker(MAX_MACRO_POINTER)
	integer macro_line_No(MAX_MACRO_POINTER)
	common /MACRO/ macro_end, macro_code_buffer, macro_code_length,
     &		macro_code_pointer, macro_cntl_marker,
     &		macro_name_end, macro_name_pointer, macro_name_buffer,
     &		macro_name_length, macro_line_No
c Symbol Buffer This is for symbol buffer.
	integer MAX_SYMBOL
	parameter (MAX_SYMBOL = 500)
	character*21 symbol_name(MAX_SYMBOL)
	integer symbol_end
	integer symbol_length(MAX_SYMBOL), symbol_value(MAX_SYMBOL)
	common /SYMBL/ symbol_end, symbol_name, symbol_length, symbol_value
c Label buffer : This is for label buffer.
	integer MAX_LABEL
	parameter (MAX_LABEL = 500)
	character*21 label_name(MAX_LABEL)
	integer label_end
	integer label_length(MAX_LABEL), gl_flag(MAX_LABEL)
	common /LBL/ label_end, label_name, label_length, gl_flag
	  macro_name_end = 0
	  macro_end = 0
	  base_end = 0
	  label_end = 0
	  symbol_end = 0
	return
	end
